import json
import os
from collections import OrderedDict

import pytest
from numpy import Infinity

import great_expectations as ge
from great_expectations.core import ExpectationSuite
from great_expectations.data_context.util import file_relative_path
from great_expectations.datasource import PandasDatasource
from great_expectations.exceptions import ProfilerError
from great_expectations.profile.basic_suite_builder_profiler import (
    BasicSuiteBuilderProfiler,
)
from tests.test_utils import expectationSuiteValidationResultSchema

FALSEY_VALUES = [None, [], False]


def test__find_next_low_card_column(
    non_numeric_low_card_dataset, non_numeric_high_card_dataset
):
    columns = non_numeric_low_card_dataset.get_table_columns()
    column_cache = {}
    profiled_columns = {"numeric": [], "low_card": [], "string": [], "datetime": []}

    column = BasicSuiteBuilderProfiler._find_next_low_card_column(
        non_numeric_low_card_dataset, columns, profiled_columns, column_cache
    )
    assert column == "lowcardnonnum"
    profiled_columns["low_card"].append(column)
    assert (
        BasicSuiteBuilderProfiler._find_next_low_card_column(
            non_numeric_low_card_dataset, columns, profiled_columns, column_cache
        )
        is None
    )

    columns = non_numeric_high_card_dataset.get_table_columns()
    column_cache = {}
    profiled_columns = {"numeric": [], "low_card": [], "string": [], "datetime": []}
    assert (
        BasicSuiteBuilderProfiler._find_next_low_card_column(
            non_numeric_high_card_dataset, columns, profiled_columns, column_cache
        )
        is None
    )


def test__create_expectations_for_low_card_column(non_numeric_low_card_dataset):
    column = "lowcardnonnum"
    column_cache = {}

    expectation_suite = non_numeric_low_card_dataset.get_expectation_suite(
        suppress_warnings=True
    )
    assert len(expectation_suite.expectations) == 1

    BasicSuiteBuilderProfiler._create_expectations_for_low_card_column(
        non_numeric_low_card_dataset, column, column_cache
    )
    expectation_suite = non_numeric_low_card_dataset.get_expectation_suite(
        suppress_warnings=True
    )
    assert set(
        [
            expectation.expectation_type
            for expectation in expectation_suite.expectations
            if expectation.kwargs.get("column") == column
        ]
    ) == {
        "expect_column_to_exist",
        "expect_column_distinct_values_to_be_in_set",
        "expect_column_kl_divergence_to_be_less_than",
        "expect_column_values_to_not_be_null",
    }


def test__find_next_numeric_column(
    numeric_high_card_dataset, non_numeric_low_card_dataset
):
    columns = numeric_high_card_dataset.get_table_columns()
    column_cache = {}
    profiled_columns = {"numeric": [], "low_card": [], "string": [], "datetime": []}

    column = BasicSuiteBuilderProfiler._find_next_numeric_column(
        numeric_high_card_dataset, columns, profiled_columns, column_cache
    )
    assert column == "norm_0_1"
    profiled_columns["numeric"].append(column)
    assert (
        BasicSuiteBuilderProfiler._find_next_numeric_column(
            numeric_high_card_dataset, columns, profiled_columns, column_cache
        )
        is None
    )

    columns = non_numeric_low_card_dataset.get_table_columns()
    column_cache = {}
    profiled_columns = {"numeric": [], "low_card": [], "string": [], "datetime": []}
    assert (
        BasicSuiteBuilderProfiler._find_next_numeric_column(
            non_numeric_low_card_dataset, columns, profiled_columns, column_cache
        )
        is None
    )


def test__create_expectations_for_numeric_column(
    numeric_high_card_dataset, test_backend
):
    column = "norm_0_1"

    expectation_suite = numeric_high_card_dataset.get_expectation_suite(
        suppress_warnings=True
    )
    assert len(expectation_suite.expectations) == 1

    BasicSuiteBuilderProfiler._create_expectations_for_numeric_column(
        numeric_high_card_dataset, column
    )
    expectation_suite = numeric_high_card_dataset.get_expectation_suite(
        suppress_warnings=True
    )
    if test_backend in ["PandasDataset", "SparkDFDataset", "postgresql"]:
        assert set(
            [
                expectation.expectation_type
                for expectation in expectation_suite.expectations
                if expectation.kwargs.get("column") == column
            ]
        ) == {
            "expect_column_to_exist",
            "expect_column_min_to_be_between",
            "expect_column_max_to_be_between",
            "expect_column_mean_to_be_between",
            "expect_column_median_to_be_between",
            "expect_column_quantile_values_to_be_between",
            "expect_column_values_to_not_be_null",
        }
    else:
        assert set(
            [
                expectation.expectation_type
                for expectation in expectation_suite.expectations
                if expectation.kwargs.get("column") == column
            ]
        ) == {
            "expect_column_to_exist",
            "expect_column_min_to_be_between",
            "expect_column_max_to_be_between",
            "expect_column_mean_to_be_between",
            "expect_column_median_to_be_between",
            "expect_column_values_to_not_be_null",
        }


def test__find_next_string_column(
    non_numeric_high_card_dataset, non_numeric_low_card_dataset
):
    columns = non_numeric_high_card_dataset.get_table_columns()
    column_cache = {}
    profiled_columns = {"numeric": [], "low_card": [], "string": [], "datetime": []}

    column = BasicSuiteBuilderProfiler._find_next_string_column(
        non_numeric_high_card_dataset, columns, profiled_columns, column_cache
    )
    expected_columns = ["highcardnonnum", "medcardnonnum"]
    assert column in expected_columns
    profiled_columns["string"].append(column)
    expected_columns.remove(column)
    assert (
        BasicSuiteBuilderProfiler._find_next_string_column(
            non_numeric_high_card_dataset, columns, profiled_columns, column_cache
        )
        in expected_columns
    )

    columns = non_numeric_low_card_dataset.get_table_columns()
    column_cache = {}
    profiled_columns = {"numeric": [], "low_card": [], "string": [], "datetime": []}
    assert (
        BasicSuiteBuilderProfiler._find_next_string_column(
            non_numeric_low_card_dataset, columns, profiled_columns, column_cache
        )
        is None
    )


def test__create_expectations_for_string_column(non_numeric_high_card_dataset):
    column = "highcardnonnum"

    expectation_suite = non_numeric_high_card_dataset.get_expectation_suite(
        suppress_warnings=True
    )
    assert len(expectation_suite.expectations) == 2

    BasicSuiteBuilderProfiler._create_expectations_for_string_column(
        non_numeric_high_card_dataset, column
    )
    expectation_suite = non_numeric_high_card_dataset.get_expectation_suite(
        suppress_warnings=True
    )
    assert set(
        [
            expectation.expectation_type
            for expectation in expectation_suite.expectations
            if expectation.kwargs.get("column") == column
        ]
    ) == {
        "expect_column_to_exist",
        "expect_column_values_to_not_be_null",
        "expect_column_value_lengths_to_be_between",
    }


def test__find_next_datetime_column(datetime_dataset, numeric_high_card_dataset):
    columns = datetime_dataset.get_table_columns()
    column_cache = {}
    profiled_columns = {"numeric": [], "low_card": [], "string": [], "datetime": []}

    column = BasicSuiteBuilderProfiler._find_next_datetime_column(
        datetime_dataset, columns, profiled_columns, column_cache
    )
    assert column == "datetime"
    profiled_columns["datetime"].append(column)
    assert (
        BasicSuiteBuilderProfiler._find_next_datetime_column(
            datetime_dataset, columns, profiled_columns, column_cache
        )
        is None
    )

    columns = numeric_high_card_dataset.get_table_columns()
    column_cache = {}
    profiled_columns = {"numeric": [], "low_card": [], "string": [], "datetime": []}
    assert (
        BasicSuiteBuilderProfiler._find_next_datetime_column(
            numeric_high_card_dataset, columns, profiled_columns, column_cache
        )
        is None
    )


def test__create_expectations_for_datetime_column(datetime_dataset):
    column = "datetime"

    expectation_suite = datetime_dataset.get_expectation_suite(suppress_warnings=True)
    assert len(expectation_suite.expectations) == 1

    BasicSuiteBuilderProfiler._create_expectations_for_datetime_column(
        datetime_dataset, column
    )
    expectation_suite = datetime_dataset.get_expectation_suite(suppress_warnings=True)
    assert set(
        [
            expectation.expectation_type
            for expectation in expectation_suite.expectations
            if expectation.kwargs.get("column") == column
        ]
    ) == {
        "expect_column_to_exist",
        "expect_column_values_to_be_between",
        "expect_column_values_to_not_be_null",
    }


def test_BasicSuiteBuilderProfiler_raises_error_on_both_included_and_excluded_expectations(
    pandas_dataset,
):
    with pytest.raises(ProfilerError):
        BasicSuiteBuilderProfiler.profile(
            pandas_dataset,
            profiler_configuration={
                "included_expectations": ["foo"],
                "excluded_expectations": ["foo"],
            },
        )


def test_BasicSuiteBuilderProfiler_raises_error_on_both_included_and_excluded_columns(
    pandas_dataset,
):
    with pytest.raises(ProfilerError):
        BasicSuiteBuilderProfiler.profile(
            pandas_dataset,
            profiler_configuration={
                "included_columns": ["foo"],
                "excluded_columns": ["foo"],
            },
        )


@pytest.mark.skipif(os.getenv("PANDAS") == "0.22.0", reason="0.22.0 pandas")
def test_BasicSuiteBuilderProfiler_raises_error_on_non_existent_column_on_pandas(
    pandas_dataset,
):
    with pytest.raises(ProfilerError):
        BasicSuiteBuilderProfiler().profile(
            pandas_dataset,
            profiler_configuration={
                "included_columns": ["NON_EXISTENT_COLUMN"],
                "included_expectations": [
                    "expect_table_column_count_to_equal",
                    "expect_column_values_to_not_be_null",
                ],
            },
        )


def test_BasicSuiteBuilderProfiler_with_context(filesystem_csv_data_context):
    context = filesystem_csv_data_context

    context.create_expectation_suite("default")
    datasource = context.datasources["rad_datasource"]
    base_dir = datasource.config["batch_kwargs_generators"]["subdir_reader"]["base_directory"]
    batch_kwargs = {
        "datasource": "rad_datasource",
        "path": os.path.join(base_dir, "f1.csv"),
    }
    batch = context.get_batch(batch_kwargs, "default")
    expectation_suite, validation_results = BasicSuiteBuilderProfiler.profile(
        batch, profiler_configuration="demo"
    )

    assert expectation_suite.expectation_suite_name == "default"
    assert "BasicSuiteBuilderProfiler" in expectation_suite.meta
    assert set(expectation_suite.meta["BasicSuiteBuilderProfiler"].keys()) == {
        "created_by",
        "created_at",
        "batch_kwargs",
    }
    assert (
        expectation_suite.meta["BasicSuiteBuilderProfiler"]["batch_kwargs"]
        == batch_kwargs
    )
    for exp in expectation_suite.expectations:
        assert "BasicSuiteBuilderProfiler" in exp.meta
        assert "confidence" in exp.meta["BasicSuiteBuilderProfiler"]

    assert set(validation_results.meta.keys()) == {
        "batch_kwargs",
        "batch_markers",
        "batch_parameters",
        "expectation_suite_name",
        "great_expectations.__version__",
        "run_id",
    }

    assert expectation_suite.meta["notes"] == {
        "format": "markdown",
        "content": [
            """#### This is an _example_ suite

- This suite was made by quickly glancing at 1000 rows of your data.
- This is **not a production suite**. It is meant to show examples of expectations.
- Because this suite was auto-generated using a very basic profiler that does not know your data like you do, many of the expectations may not be meaningful.
"""
        ],
    }

    expectation_types = [
        expectation["expectation_type"]
        for expectation in expectation_suite.expectations
    ]

    expected_expectation_types = {
        "expect_table_row_count_to_be_between",
        "expect_table_column_count_to_equal",
        "expect_table_columns_to_match_ordered_list",
        "expect_column_values_to_not_be_null",
        "expect_column_min_to_be_between",
        "expect_column_max_to_be_between",
        "expect_column_mean_to_be_between",
        "expect_column_median_to_be_between",
        "expect_column_quantile_values_to_be_between",
    }

    assert set(expectation_types) == expected_expectation_types


def test_context_profiler(filesystem_csv_data_context):
    """
    This just validates that it's possible to profile using the datasource hook,
    and have validation results available in the DataContext
    """
    context = filesystem_csv_data_context

    assert isinstance(context.datasources["rad_datasource"], PandasDatasource)
    assert context.list_expectation_suites() == []
    context.profile_datasource(
        "rad_datasource",
        profiler=BasicSuiteBuilderProfiler,
        profiler_configuration="demo",
    )

    assert len(context.list_expectation_suites()) == 1

    expected_suite_name = "rad_datasource.subdir_reader.f1.BasicSuiteBuilderProfiler"
    expectation_suite = context.get_expectation_suite(expected_suite_name)

    for exp in expectation_suite.expectations:
        assert "BasicSuiteBuilderProfiler" in exp.meta
        assert "confidence" in exp.meta["BasicSuiteBuilderProfiler"]

    assert expectation_suite.expectation_suite_name == expected_suite_name
    assert "batch_kwargs" in expectation_suite.meta["BasicSuiteBuilderProfiler"]

    assert expectation_suite.meta["notes"] == {
        "format": "markdown",
        "content": [
            """#### This is an _example_ suite

- This suite was made by quickly glancing at 1000 rows of your data.
- This is **not a production suite**. It is meant to show examples of expectations.
- Because this suite was auto-generated using a very basic profiler that does not know your data like you do, many of the expectations may not be meaningful.
"""
        ],
    }

    expectation_types = [
        expectation["expectation_type"]
        for expectation in expectation_suite.expectations
    ]

    expected_expectation_types = {
        "expect_table_row_count_to_be_between",
        "expect_table_column_count_to_equal",
        "expect_table_columns_to_match_ordered_list",
        "expect_column_values_to_not_be_null",
        "expect_column_min_to_be_between",
        "expect_column_max_to_be_between",
        "expect_column_mean_to_be_between",
        "expect_column_median_to_be_between",
        "expect_column_quantile_values_to_be_between",
    }

    assert set(expectation_types) == expected_expectation_types


def test_snapshot_BasicSuiteBuilderProfiler_on_titanic_in_demo_mode():
    """
    A snapshot regression test for BasicSuiteBuilderProfiler.
    We are running the profiler on the Titanic dataset
    and comparing the EVRs to ones retrieved from a
    previously stored file.
    """
    df = ge.read_csv(file_relative_path(__file__, "../test_sets/Titanic.csv"))
    suite, evrs = df.profile(BasicSuiteBuilderProfiler, profiler_configuration="demo")

    # Check to make sure BasicSuiteBuilderProfiler is adding meta.columns with a single "description" field for each column
    assert "columns" in suite.meta
    for k, v in suite.meta["columns"].items():
        assert v == {"description": ""}

    # Note: the above already produces an EVR; rerunning isn't strictly necessary just for EVRs
    evrs = df.validate(result_format="SUMMARY")

    # THIS IS NOT DEAD CODE. UNCOMMENT TO SAVE A SNAPSHOT WHEN UPDATING THIS TEST
    expected_filepath = file_relative_path(
        __file__,
        "./fixtures/expected_evrs_BasicSuiteBuilderProfiler_on_titanic_demo_mode.json",
    )
    # with open(expected_filepath, 'w+') as file:
    #     json.dump(expectationSuiteValidationResultSchema.dump(evrs), file, indent=2)
    # with open(file_relative_path(__file__, '../render/fixtures/BasicSuiteBuilderProfiler_evrs.json'), 'w+') as file:
    #     json.dump(expectationSuiteValidationResultSchema.dump(evrs), file, indent=2)

    with open(expected_filepath, "r",) as file:
        expected_evrs = expectationSuiteValidationResultSchema.load(
            json.load(file, object_pairs_hook=OrderedDict)
        )

    # We know that python 2 does not guarantee the order of value_counts, which causes a different
    # order for items in the partial_unexpected_value_counts list
    # Remove those before assertions.
    for result in evrs.results:
        if "partial_unexpected_counts" in result.result:
            result.result.pop("partial_unexpected_counts")

    for result in expected_evrs.results:
        if "partial_unexpected_counts" in result.result:
            result.result.pop("partial_unexpected_counts")

    # Version and RUN-ID will be different
    del expected_evrs.meta["great_expectations.__version__"]
    del evrs.meta["great_expectations.__version__"]
    del expected_evrs.meta["run_id"]
    del expected_evrs.meta["batch_kwargs"]["ge_batch_id"]
    del evrs.meta["run_id"]
    del evrs.meta["batch_kwargs"]["ge_batch_id"]

    assert expected_evrs == evrs


@pytest.mark.skipif(os.getenv("PANDAS") == "0.22.0", reason="0.22.0 pandas")
def test_BasicSuiteBuilderProfiler_uses_all_columns_if_configuration_does_not_have_included_or_excluded_columns_on_pandas(
    pandas_dataset,
):
    observed_suite, evrs = BasicSuiteBuilderProfiler().profile(pandas_dataset)
    assert isinstance(observed_suite, ExpectationSuite)

    expected = ExpectationSuite(
        "default",
        data_asset_type="Dataset",
        expectations=[
            {
                "kwargs": {"column": "infinities"},
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_column_to_exist",
            },
            {
                "kwargs": {"column": "nulls"},
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_column_to_exist",
            },
            {
                "kwargs": {"column": "naturals"},
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_column_to_exist",
            },
            {
                "kwargs": {"min_value": 6, "max_value": 7},
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_table_row_count_to_be_between",
            },
            {
                "kwargs": {"value": 3},
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_table_column_count_to_equal",
            },
            {
                "kwargs": {"column_list": ["infinities", "nulls", "naturals"]},
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_table_columns_to_match_ordered_list",
            },
            {
                "kwargs": {"column": "infinities"},
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_column_values_to_be_unique",
            },
            {
                "kwargs": {"column": "infinities"},
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_column_values_to_not_be_null",
            },
            {
                "kwargs": {
                    "column": "infinities",
                    "min_value": -Infinity,
                    "max_value": -Infinity,
                },
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_column_min_to_be_between",
            },
            {
                "kwargs": {
                    "column": "infinities",
                    "min_value": Infinity,
                    "max_value": Infinity,
                },
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_column_max_to_be_between",
            },
            {
                "kwargs": {
                    "column": "infinities",
                    "min_value": None,
                    "max_value": None,
                },
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_column_mean_to_be_between",
            },
            {
                "kwargs": {"column": "infinities", "min_value": -1.0, "max_value": 1.0},
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_column_median_to_be_between",
            },
            {
                "kwargs": {
                    "column": "infinities",
                    "quantile_ranges": {
                        "quantiles": [0.05, 0.25, 0.5, 0.75, 0.95],
                        "value_ranges": [
                            [-Infinity, -Infinity],
                            [-4.141592653589793, -2.141592653589793],
                            [-1.0, 1.0],
                            [2.141592653589793, 4.141592653589793],
                            [Infinity, Infinity],
                        ],
                    },
                },
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_column_quantile_values_to_be_between",
            },
            {
                "kwargs": {"column": "nulls"},
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_column_values_to_be_unique",
            },
            {
                "kwargs": {"column": "nulls", "mostly": 0.4714285714285715},
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_column_values_to_not_be_null",
            },
            {
                "kwargs": {"column": "nulls", "min_value": -1.0, "max_value": 1.0},
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_column_min_to_be_between",
            },
            {
                "kwargs": {"column": "nulls", "min_value": 2.3, "max_value": 4.3},
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_column_max_to_be_between",
            },
            {
                "kwargs": {
                    "column": "nulls",
                    "min_value": 0.6499999999999999,
                    "max_value": 2.65,
                },
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_column_mean_to_be_between",
            },
            {
                "kwargs": {
                    "column": "nulls",
                    "min_value": 0.6500000000000001,
                    "max_value": 2.6500000000000004,
                },
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_column_median_to_be_between",
            },
            {
                "kwargs": {
                    "column": "nulls",
                    "quantile_ranges": {
                        "quantiles": [0.05, 0.25, 0.5, 0.75, 0.95],
                        "value_ranges": [
                            [-1.0, 1.0],
                            [0.10000000000000009, 2.1],
                            [1.2000000000000002, 3.2],
                            [1.2000000000000002, 3.2],
                            [2.3, 4.3],
                        ],
                    },
                },
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_column_quantile_values_to_be_between",
            },
            {
                "kwargs": {"column": "naturals"},
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_column_values_to_be_unique",
            },
            {
                "kwargs": {"column": "naturals"},
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_column_values_to_not_be_null",
            },
            {
                "kwargs": {"column": "naturals", "min_value": 0.0, "max_value": 2.0},
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_column_min_to_be_between",
            },
            {
                "kwargs": {"column": "naturals", "min_value": 6.0, "max_value": 8.0},
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_column_max_to_be_between",
            },
            {
                "kwargs": {"column": "naturals", "min_value": 3.0, "max_value": 5.0},
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_column_mean_to_be_between",
            },
            {
                "kwargs": {"column": "naturals", "min_value": 3.0, "max_value": 5.0},
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_column_median_to_be_between",
            },
            {
                "kwargs": {
                    "column": "naturals",
                    "quantile_ranges": {
                        "quantiles": [0.05, 0.25, 0.5, 0.75, 0.95],
                        "value_ranges": [
                            [0.0, 2.0],
                            [2.0, 4.0],
                            [3.0, 5.0],
                            [4.0, 6.0],
                            [6.0, 8.0],
                        ],
                    },
                },
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_column_quantile_values_to_be_between",
            },
        ],
    )

    # remove metadata to simplify assertions
    observed_suite.meta = None
    expected.meta = None
    assert observed_suite == expected


@pytest.mark.skipif(os.getenv("PANDAS") == "0.22.0", reason="0.22.0 pandas")
def test_BasicSuiteBuilderProfiler_uses_selected_columns_on_pandas(pandas_dataset,):
    observed_suite, evrs = BasicSuiteBuilderProfiler().profile(
        pandas_dataset, profiler_configuration={"included_columns": ["naturals"]}
    )
    assert isinstance(observed_suite, ExpectationSuite)

    expected = ExpectationSuite(
        "default",
        data_asset_type="Dataset",
        expectations=[
            {
                "kwargs": {"column": "infinities"},
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_column_to_exist",
            },
            {
                "kwargs": {"column": "nulls"},
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_column_to_exist",
            },
            {
                "kwargs": {"column": "naturals"},
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_column_to_exist",
            },
            {
                "kwargs": {"min_value": 6, "max_value": 7},
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_table_row_count_to_be_between",
            },
            {
                "kwargs": {"value": 3},
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_table_column_count_to_equal",
            },
            {
                "kwargs": {"column_list": ["infinities", "nulls", "naturals"]},
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_table_columns_to_match_ordered_list",
            },
            {
                "kwargs": {"column": "naturals"},
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_column_values_to_be_unique",
            },
            {
                "kwargs": {"column": "naturals"},
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_column_values_to_not_be_null",
            },
            {
                "kwargs": {"column": "naturals", "min_value": 0.0, "max_value": 2.0},
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_column_min_to_be_between",
            },
            {
                "kwargs": {"column": "naturals", "min_value": 6.0, "max_value": 8.0},
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_column_max_to_be_between",
            },
            {
                "kwargs": {"column": "naturals", "min_value": 3.0, "max_value": 5.0},
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_column_mean_to_be_between",
            },
            {
                "kwargs": {"column": "naturals", "min_value": 3.0, "max_value": 5.0},
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_column_median_to_be_between",
            },
            {
                "kwargs": {
                    "column": "naturals",
                    "quantile_ranges": {
                        "quantiles": [0.05, 0.25, 0.5, 0.75, 0.95],
                        "value_ranges": [
                            [0.0, 2.0],
                            [2.0, 4.0],
                            [3.0, 5.0],
                            [4.0, 6.0],
                            [6.0, 8.0],
                        ],
                    },
                },
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "expectation_type": "expect_column_quantile_values_to_be_between",
            },
        ],
    )

    # remove metadata to simplify assertions
    observed_suite.meta = None
    expected.meta = None
    assert observed_suite == expected


@pytest.mark.skipif(os.getenv("PANDAS") == "0.22.0", reason="0.22.0 pandas")
def test_BasicSuiteBuilderProfiler_respects_excluded_expectations_on_pandas(
    pandas_dataset,
):
    observed_suite, evrs = BasicSuiteBuilderProfiler().profile(
        pandas_dataset,
        profiler_configuration={
            "included_columns": ["naturals"],
            "excluded_expectations": [
                "expect_table_column_count_to_equal",
                "expect_column_values_to_be_unique",
                "expect_column_min_to_be_between",
                "expect_column_max_to_be_between",
                "expect_column_mean_to_be_between",
                "expect_column_median_to_be_between",
                "expect_column_quantile_values_to_be_between",
            ],
        },
    )
    assert isinstance(observed_suite, ExpectationSuite)

    expected = ExpectationSuite(
        "default",
        data_asset_type="Dataset",
        expectations=[
            {
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "kwargs": {"column": "infinities"},
                "expectation_type": "expect_column_to_exist",
            },
            {
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "kwargs": {"column": "nulls"},
                "expectation_type": "expect_column_to_exist",
            },
            {
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "kwargs": {"column": "naturals"},
                "expectation_type": "expect_column_to_exist",
            },
            {
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "kwargs": {"min_value": 6, "max_value": 7},
                "expectation_type": "expect_table_row_count_to_be_between",
            },
            {
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "kwargs": {"column_list": ["infinities", "nulls", "naturals"]},
                "expectation_type": "expect_table_columns_to_match_ordered_list",
            },
            {
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "kwargs": {"column": "naturals"},
                "expectation_type": "expect_column_values_to_not_be_null",
            },
        ],
    )

    # remove metadata to simplify assertions
    observed_suite.meta = None
    expected.meta = None
    assert observed_suite == expected


@pytest.mark.skipif(os.getenv("PANDAS") == "0.22.0", reason="0.22.0 pandas")
def test_BasicSuiteBuilderProfiler_respects_included_expectations_on_pandas(
    pandas_dataset,
):
    observed_suite, evrs = BasicSuiteBuilderProfiler().profile(
        pandas_dataset,
        profiler_configuration={
            "included_columns": ["naturals", "nulls"],
            "included_expectations": [
                "expect_table_column_count_to_equal",
                "expect_column_values_to_not_be_null",
            ],
        },
    )
    assert isinstance(observed_suite, ExpectationSuite)

    expected = ExpectationSuite(
        "default",
        data_asset_type="Dataset",
        expectations=[
            {
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "kwargs": {"value": 3},
                "expectation_type": "expect_table_column_count_to_equal",
            },
            {
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "kwargs": {"column": "naturals"},
                "expectation_type": "expect_column_values_to_not_be_null",
            },
            {
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "kwargs": {"column": "nulls", "mostly": 0.4714285714285715},
                "expectation_type": "expect_column_values_to_not_be_null",
            },
        ],
    )

    # remove metadata to simplify assertions
    observed_suite.meta = None
    expected.meta = None
    assert observed_suite == expected


@pytest.mark.skipif(os.getenv("PANDAS") == "0.22.0", reason="0.22.0 pandas")
@pytest.mark.parametrize("included_columns", FALSEY_VALUES)
def test_BasicSuiteBuilderProfiler_uses_no_columns_if_included_columns_are_falsey_on_pandas(
    included_columns, pandas_dataset,
):
    observed_suite, evrs = BasicSuiteBuilderProfiler().profile(
        pandas_dataset,
        profiler_configuration={
            "included_columns": included_columns,
            "included_expectations": [
                "expect_table_column_count_to_equal",
                "expect_column_values_to_not_be_null",
            ],
        },
    )
    assert isinstance(observed_suite, ExpectationSuite)

    expected = ExpectationSuite(
        "default",
        data_asset_type="Dataset",
        expectations=[
            {
                "expectation_type": "expect_table_column_count_to_equal",
                "kwargs": {"value": 3},
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
            },
        ],
    )

    # remove metadata to simplify assertions
    observed_suite.meta = None
    expected.meta = None
    assert observed_suite == expected


@pytest.mark.skipif(os.getenv("PANDAS") == "0.22.0", reason="0.22.0 pandas")
@pytest.mark.parametrize("included_expectations", FALSEY_VALUES)
def test_BasicSuiteBuilderProfiler_uses_no_expectations_if_included_expectations_are_falsey_on_pandas(
    included_expectations, pandas_dataset,
):
    observed_suite, evrs = BasicSuiteBuilderProfiler().profile(
        pandas_dataset,
        profiler_configuration={"included_expectations": included_expectations,},
    )
    assert isinstance(observed_suite, ExpectationSuite)

    expected = ExpectationSuite("default", data_asset_type="Dataset", expectations=[],)

    # remove metadata to simplify assertions
    observed_suite.meta = None
    expected.meta = None
    assert observed_suite == expected


@pytest.mark.skipif(os.getenv("PANDAS") == "0.22.0", reason="0.22.0 pandas")
@pytest.mark.parametrize("excluded_expectations", FALSEY_VALUES)
def test_BasicSuiteBuilderProfiler_uses_all_expectations_if_excluded_expectations_are_falsey_on_pandas(
    excluded_expectations, pandas_dataset,
):
    observed_suite, evrs = BasicSuiteBuilderProfiler().profile(
        pandas_dataset,
        profiler_configuration={"excluded_expectations": excluded_expectations,},
    )
    assert isinstance(observed_suite, ExpectationSuite)

    expected = ExpectationSuite(
        "default",
        data_asset_type="Dataset",
        expectations=[
            {
                "kwargs": {"column": "infinities"},
                "expectation_type": "expect_column_to_exist",
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
            },
            {
                "kwargs": {"column": "nulls"},
                "expectation_type": "expect_column_to_exist",
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
            },
            {
                "kwargs": {"column": "naturals"},
                "expectation_type": "expect_column_to_exist",
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
            },
            {
                "kwargs": {"min_value": 6, "max_value": 7},
                "expectation_type": "expect_table_row_count_to_be_between",
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
            },
            {
                "kwargs": {"value": 3},
                "expectation_type": "expect_table_column_count_to_equal",
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
            },
            {
                "kwargs": {"column_list": ["infinities", "nulls", "naturals"]},
                "expectation_type": "expect_table_columns_to_match_ordered_list",
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
            },
            {
                "kwargs": {"column": "infinities"},
                "expectation_type": "expect_column_values_to_be_unique",
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
            },
            {
                "kwargs": {"column": "infinities"},
                "expectation_type": "expect_column_values_to_not_be_null",
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
            },
            {
                "kwargs": {
                    "column": "infinities",
                    "min_value": -Infinity,
                    "max_value": -Infinity,
                },
                "expectation_type": "expect_column_min_to_be_between",
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
            },
            {
                "kwargs": {
                    "column": "infinities",
                    "min_value": Infinity,
                    "max_value": Infinity,
                },
                "expectation_type": "expect_column_max_to_be_between",
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
            },
            {
                "kwargs": {
                    "column": "infinities",
                    "min_value": None,
                    "max_value": None,
                },
                "expectation_type": "expect_column_mean_to_be_between",
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
            },
            {
                "kwargs": {"column": "infinities", "min_value": -1.0, "max_value": 1.0},
                "expectation_type": "expect_column_median_to_be_between",
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
            },
            {
                "kwargs": {
                    "column": "infinities",
                    "quantile_ranges": {
                        "quantiles": [0.05, 0.25, 0.5, 0.75, 0.95],
                        "value_ranges": [
                            [-Infinity, -Infinity],
                            [-4.141592653589793, -2.141592653589793],
                            [-1.0, 1.0],
                            [2.141592653589793, 4.141592653589793],
                            [Infinity, Infinity],
                        ],
                    },
                },
                "expectation_type": "expect_column_quantile_values_to_be_between",
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
            },
            {
                "kwargs": {"column": "nulls"},
                "expectation_type": "expect_column_values_to_be_unique",
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
            },
            {
                "kwargs": {"column": "nulls", "mostly": 0.4714285714285715},
                "expectation_type": "expect_column_values_to_not_be_null",
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
            },
            {
                "kwargs": {"column": "nulls", "min_value": -1.0, "max_value": 1.0},
                "expectation_type": "expect_column_min_to_be_between",
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
            },
            {
                "kwargs": {"column": "nulls", "min_value": 2.3, "max_value": 4.3},
                "expectation_type": "expect_column_max_to_be_between",
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
            },
            {
                "kwargs": {
                    "column": "nulls",
                    "min_value": 0.6499999999999999,
                    "max_value": 2.65,
                },
                "expectation_type": "expect_column_mean_to_be_between",
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
            },
            {
                "kwargs": {
                    "column": "nulls",
                    "min_value": 0.6500000000000001,
                    "max_value": 2.6500000000000004,
                },
                "expectation_type": "expect_column_median_to_be_between",
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
            },
            {
                "kwargs": {
                    "column": "nulls",
                    "quantile_ranges": {
                        "quantiles": [0.05, 0.25, 0.5, 0.75, 0.95],
                        "value_ranges": [
                            [-1.0, 1.0],
                            [0.10000000000000009, 2.1],
                            [1.2000000000000002, 3.2],
                            [1.2000000000000002, 3.2],
                            [2.3, 4.3],
                        ],
                    },
                },
                "expectation_type": "expect_column_quantile_values_to_be_between",
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
            },
            {
                "kwargs": {"column": "naturals"},
                "expectation_type": "expect_column_values_to_be_unique",
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
            },
            {
                "kwargs": {"column": "naturals"},
                "expectation_type": "expect_column_values_to_not_be_null",
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
            },
            {
                "kwargs": {"column": "naturals", "min_value": 0.0, "max_value": 2.0},
                "expectation_type": "expect_column_min_to_be_between",
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
            },
            {
                "kwargs": {"column": "naturals", "min_value": 6.0, "max_value": 8.0},
                "expectation_type": "expect_column_max_to_be_between",
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
            },
            {
                "kwargs": {"column": "naturals", "min_value": 3.0, "max_value": 5.0},
                "expectation_type": "expect_column_mean_to_be_between",
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
            },
            {
                "kwargs": {"column": "naturals", "min_value": 3.0, "max_value": 5.0},
                "expectation_type": "expect_column_median_to_be_between",
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
            },
            {
                "kwargs": {
                    "column": "naturals",
                    "quantile_ranges": {
                        "quantiles": [0.05, 0.25, 0.5, 0.75, 0.95],
                        "value_ranges": [
                            [0.0, 2.0],
                            [2.0, 4.0],
                            [3.0, 5.0],
                            [4.0, 6.0],
                            [6.0, 8.0],
                        ],
                    },
                },
                "expectation_type": "expect_column_quantile_values_to_be_between",
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
            },
        ],
    )

    # remove metadata to simplify assertions
    observed_suite.meta = None
    expected.meta = None
    assert observed_suite == expected


@pytest.mark.skipif(os.getenv("PANDAS") == "0.22.0", reason="0.22.0 pandas")
@pytest.mark.parametrize("excluded_columns", FALSEY_VALUES)
def test_BasicSuiteBuilderProfiler_uses_all_columns_if_excluded_columns_are_falsey_on_pandas(
    excluded_columns, pandas_dataset,
):
    observed_suite, evrs = BasicSuiteBuilderProfiler().profile(
        pandas_dataset,
        profiler_configuration={
            "excluded_columns": excluded_columns,
            "included_expectations": [
                "expect_table_column_count_to_equal",
                "expect_column_values_to_not_be_null",
            ],
        },
    )
    assert isinstance(observed_suite, ExpectationSuite)

    expected = ExpectationSuite(
        "default",
        data_asset_type="Dataset",
        expectations=[
            {
                "expectation_type": "expect_table_column_count_to_equal",
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "kwargs": {"value": 3},
            },
            {
                "expectation_type": "expect_column_values_to_not_be_null",
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "kwargs": {"column": "naturals"},
            },
            {
                "expectation_type": "expect_column_values_to_not_be_null",
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "kwargs": {"column": "infinities"},
            },
            {
                "expectation_type": "expect_column_values_to_not_be_null",
                "meta": {"BasicSuiteBuilderProfiler": {"confidence": "very low"}},
                "kwargs": {"column": "nulls", "mostly": 0.4714285714285715},
            },
        ],
    )

    # remove metadata to simplify assertions
    observed_suite.meta = None
    expected.meta = None
    assert len(observed_suite.expectations) == 4
    for ee in expected.expectations:
        assert ee in observed_suite.expectations
    # assert observed_suite == expected


@pytest.mark.skipif(os.getenv("PANDAS") == "0.22.0", reason="0.22.0 pandas")
def test_BasicSuiteBuilderProfiler_raises_error_on_not_real_expectations_in_included_expectations_on_pandas(
    pandas_dataset,
):
    with pytest.raises(ProfilerError):
        BasicSuiteBuilderProfiler().profile(
            pandas_dataset,
            profiler_configuration={
                "included_expectations": [
                    "expect_table_column_count_to_equal",
                    "expect_this_to_not_be_a_real_expectation",
                    "foo",
                ],
            },
        )


@pytest.mark.skipif(os.getenv("PANDAS") == "0.22.0", reason="0.22.0 pandas")
def test_BasicSuiteBuilderProfiler_raises_error_on_not_real_expectations_in_excluded_expectations_on_pandas(
    pandas_dataset,
):
    with pytest.raises(ProfilerError):
        BasicSuiteBuilderProfiler().profile(
            pandas_dataset,
            profiler_configuration={
                "included_expectations": [
                    "expect_table_column_count_to_equal",
                    "expect_this_to_not_be_a_real_expectation",
                    "foo",
                ],
            },
        )


def test_snapshot_BasicSuiteBuilderProfiler_on_titanic_with_builder_configuration():
    """
    A snapshot regression test for BasicSuiteBuilderProfiler.

    We are running the profiler on the Titanic dataset and comparing the EVRs
    to ones retrieved from a previously stored file.
    """
    batch = ge.read_csv(file_relative_path(__file__, "../test_sets/Titanic.csv"))
    suite, evrs = BasicSuiteBuilderProfiler().profile(
        batch,
        profiler_configuration={
            "included_columns": ["Name", "PClass", "Age", "Sex", "SexCode"]
        },
    )

    # Check to make sure SuiteBuilderProfiler is adding meta.columns with a single "description" field for each column
    assert "columns" in suite.meta
    for k, v in suite.meta["columns"].items():
        assert v == {"description": ""}

    # Note: the above already produces an EVR; rerunning isn't strictly necessary just for EVRs
    evrs = batch.validate(result_format="SUMMARY")

    expected_filepath = file_relative_path(
        __file__,
        "./fixtures/expected_evrs_SuiteBuilderProfiler_on_titanic_with_configurations.json",
    )
    # THIS IS NOT DEAD CODE. UNCOMMENT TO SAVE A SNAPSHOT WHEN UPDATING THIS TEST
    # with open(expected_filepath, 'w+') as file:
    #     json.dump(expectationSuiteValidationResultSchema.dump(evrs), file, indent=2)
    # with open(file_relative_path(__file__, '../render/fixtures/SuiteBuilderProfiler_evrs.json'), 'w+') as file:
    #     json.dump(expectationSuiteValidationResultSchema.dump(evrs), file, indent=2)

    with open(expected_filepath, "r",) as file:
        expected_evrs = expectationSuiteValidationResultSchema.load(
            json.load(file, object_pairs_hook=OrderedDict)
        )

        # Version and RUN-ID will be different
    del expected_evrs.meta["great_expectations.__version__"]
    del evrs.meta["great_expectations.__version__"]
    del expected_evrs.meta["run_id"]
    del expected_evrs.meta["batch_kwargs"]["ge_batch_id"]
    del evrs.meta["run_id"]
    del evrs.meta["batch_kwargs"]["ge_batch_id"]

    assert expected_evrs == evrs



import datetime
import os
import subprocess
import sys
from typing import Union

import click

from great_expectations import DataContext
from great_expectations import exceptions as ge_exceptions
from great_expectations.cli.datasource import get_batch_kwargs
from great_expectations.cli.docs import build_docs
from great_expectations.cli.util import cli_message
from great_expectations.core import ExpectationSuite
from great_expectations.core.id_dict import BatchKwargs
from great_expectations.core.usage_statistics.usage_statistics import (
    send_usage_message,
)
from great_expectations.data_asset import DataAsset
from great_expectations.data_context.types.resource_identifiers import (
    ValidationResultIdentifier,
)
from great_expectations.datasource import Datasource
from great_expectations.exceptions import (
    CheckpointError,
    CheckpointNotFoundError,
)
from great_expectations.profile import BasicSuiteBuilderProfiler


def create_expectation_suite(
    context,
    datasource_name=None,
    batch_kwargs_generator_name=None,
    generator_asset=None,
    batch_kwargs=None,
    expectation_suite_name=None,
    additional_batch_kwargs=None,
    empty_suite=False,
    show_intro_message=False,
    open_docs=False,
    profiler_configuration="demo",
):
    """
    Create a new expectation suite.

    :return: a tuple: (success, suite name)
    """
    if show_intro_message and not empty_suite:
        cli_message(
            "\n<cyan>========== Create sample Expectations ==========</cyan>\n\n"
        )

    data_source = select_datasource(context, datasource_name=datasource_name)
    if data_source is None:
        # select_datasource takes care of displaying an error message, so all is left here is to exit.
        sys.exit(1)

    datasource_name = data_source.name

    if expectation_suite_name in context.list_expectation_suite_names():
        tell_user_suite_exists(expectation_suite_name)
        sys.exit(1)

    if (
        batch_kwargs_generator_name is None
        or generator_asset is None
        or batch_kwargs is None
    ):
        (
            datasource_name,
            batch_kwargs_generator_name,
            generator_asset,
            batch_kwargs,
        ) = get_batch_kwargs(
            context,
            datasource_name=datasource_name,
            batch_kwargs_generator_name=batch_kwargs_generator_name,
            generator_asset=generator_asset,
            additional_batch_kwargs=additional_batch_kwargs,
        )
        # In this case, we have "consumed" the additional_batch_kwargs
        additional_batch_kwargs = {}

    if expectation_suite_name is None:
        default_expectation_suite_name = _get_default_expectation_suite_name(
            batch_kwargs, generator_asset
        )
        while True:
            expectation_suite_name = click.prompt(
                "\nName the new expectation suite",
                default=default_expectation_suite_name,
            )
            if expectation_suite_name in context.list_expectation_suite_names():
                tell_user_suite_exists(expectation_suite_name)
            else:
                break

    if empty_suite:
        create_empty_suite(context, expectation_suite_name, batch_kwargs)
        return True, expectation_suite_name

    profiling_results = _profile_to_create_a_suite(
        additional_batch_kwargs,
        batch_kwargs,
        batch_kwargs_generator_name,
        context,
        datasource_name,
        expectation_suite_name,
        generator_asset,
        profiler_configuration,
    )

    build_docs(context, view=False)
    if open_docs:
        _attempt_to_open_validation_results_in_data_docs(context, profiling_results)

    return True, expectation_suite_name


def _profile_to_create_a_suite(
    additional_batch_kwargs,
    batch_kwargs,
    batch_kwargs_generator_name,
    context,
    datasource_name,
    expectation_suite_name,
    generator_asset,
    profiler_configuration,
):
    click.prompt(
        """
Great Expectations will choose a couple of columns and generate expectations about them
to demonstrate some examples of assertions you can make about your data. 
    
Press Enter to continue
""",
        default=True,
        show_default=False,
    )
    # TODO this may not apply
    cli_message("\nGenerating example Expectation Suite...")
    run_id = datetime.datetime.utcnow().strftime("%Y%m%dT%H%M%S.%fZ")
    profiling_results = context.profile_data_asset(
        datasource_name,
        batch_kwargs_generator_name=batch_kwargs_generator_name,
        data_asset_name=generator_asset,
        batch_kwargs=batch_kwargs,
        profiler=BasicSuiteBuilderProfiler,
        profiler_configuration=profiler_configuration,
        expectation_suite_name=expectation_suite_name,
        run_id=run_id,
        additional_batch_kwargs=additional_batch_kwargs,
    )
    if not profiling_results["success"]:
        _raise_profiling_errors(profiling_results)
    return profiling_results


def _raise_profiling_errors(profiling_results):
    if (
        profiling_results["error"]["code"]
        == DataContext.PROFILING_ERROR_CODE_SPECIFIED_DATA_ASSETS_NOT_FOUND
    ):
        raise ge_exceptions.DataContextError(
            """Some of the data assets you specified were not found: {0:s}    
            """.format(
                ",".join(profiling_results["error"]["not_found_data_assets"])
            )
        )
    raise ge_exceptions.DataContextError(
        "Unknown profiling error code: " + profiling_results["error"]["code"]
    )


def _attempt_to_open_validation_results_in_data_docs(context, profiling_results):
    try:
        # TODO this is really brittle and not covered in tests
        validation_result = profiling_results["results"][0][1]
        validation_result_identifier = ValidationResultIdentifier.from_object(
            validation_result
        )
        context.open_data_docs(resource_identifier=validation_result_identifier)
    except (KeyError, IndexError):
        context.open_data_docs()


def _get_default_expectation_suite_name(batch_kwargs, generator_asset):
    if generator_asset:
        suite_name = f"{generator_asset}.warning"
    elif "query" in batch_kwargs:
        suite_name = "query.warning"
    elif "path" in batch_kwargs:
        try:
            # Try guessing a filename
            filename = os.path.split(os.path.normpath(batch_kwargs["path"]))[1]
            # Take all but the last part after the period
            filename = ".".join(filename.split(".")[:-1])
            suite_name = str(filename) + ".warning"
        except (OSError, IndexError):
            suite_name = "warning"
    else:
        suite_name = "warning"
    return suite_name


def tell_user_suite_exists(suite_name: str) -> None:
    cli_message(
        f"""<red>An expectation suite named `{suite_name}` already exists.</red>
  - If you intend to edit the suite please use `great_expectations suite edit {suite_name}`."""
    )


def create_empty_suite(
    context: DataContext, expectation_suite_name: str, batch_kwargs
) -> None:
    suite = context.create_expectation_suite(expectation_suite_name)
    suite.add_citation(comment="New suite added via CLI", batch_kwargs=batch_kwargs)
    context.save_expectation_suite(suite, expectation_suite_name)


def launch_jupyter_notebook(notebook_path: str) -> None:
    subprocess.call(["jupyter", "notebook", notebook_path])


def load_batch(
    context: DataContext,
    suite: Union[str, ExpectationSuite],
    batch_kwargs: Union[dict, BatchKwargs],
) -> DataAsset:
    batch: DataAsset = context.get_batch(batch_kwargs, suite)
    assert isinstance(
        batch, DataAsset
    ), "Batch failed to load. Please check your batch_kwargs"
    return batch


def load_expectation_suite(
    # TODO consolidate all the myriad CLI tests into this
    context: DataContext,
    suite_name: str,
    usage_event: str,
) -> ExpectationSuite:
    """
    Load an expectation suite from a given context.

    Handles a suite name with or without `.json`
    :param usage_event:
    """
    if suite_name.endswith(".json"):
        suite_name = suite_name[:-5]
    try:
        suite = context.get_expectation_suite(suite_name)
        return suite
    except ge_exceptions.DataContextError as e:
        exit_with_failure_message_and_stats(
            context,
            usage_event,
            f"<red>Could not find a suite named `{suite_name}`.</red> Please check "
            "the name by running `great_expectations suite list` and try again.",
        )


def exit_with_failure_message_and_stats(
    context: DataContext, usage_event: str, message: str
) -> None:
    cli_message(message)
    send_usage_message(context, event=usage_event, success=False)
    sys.exit(1)


def load_checkpoint(
    context: DataContext, checkpoint_name: str, usage_event: str
) -> dict:
    """Load a checkpoint or raise helpful errors."""
    try:
        checkpoint_config = context.get_checkpoint(checkpoint_name)
        return checkpoint_config
    except CheckpointNotFoundError as e:
        exit_with_failure_message_and_stats(
            context,
            usage_event,
            f"""\
<red>Could not find checkpoint `{checkpoint_name}`.</red> Try running:
  - `<green>great_expectations checkpoint list</green>` to verify your checkpoint exists
  - `<green>great_expectations checkpoint new</green>` to configure a new checkpoint""",
        )
    except CheckpointError as e:
        exit_with_failure_message_and_stats(context, usage_event, f"<red>{e}</red>")


def select_datasource(context: DataContext, datasource_name: str = None) -> Datasource:
    """Select a datasource interactively."""
    # TODO consolidate all the myriad CLI tests into this
    data_source = None

    if datasource_name is None:
        data_sources = sorted(context.list_datasources(), key=lambda x: x["name"])
        if len(data_sources) == 0:
            cli_message(
                "<red>No datasources found in the context. To add a datasource, run `great_expectations datasource new`</red>"
            )
        elif len(data_sources) == 1:
            datasource_name = data_sources[0]["name"]
        else:
            choices = "\n".join(
                [
                    "    {}. {}".format(i, data_source["name"])
                    for i, data_source in enumerate(data_sources, 1)
                ]
            )
            option_selection = click.prompt(
                "Select a datasource" + "\n" + choices + "\n",
                type=click.Choice(
                    [str(i) for i, data_source in enumerate(data_sources, 1)]
                ),
                show_choices=False,
            )
            datasource_name = data_sources[int(option_selection) - 1]["name"]

    if datasource_name is not None:
        data_source = context.get_datasource(datasource_name)

    return data_source


def load_data_context_with_error_handling(directory: str) -> DataContext:
    """Return a DataContext with good error handling and exit codes."""
    # TODO consolidate all the myriad CLI tests into this
    try:
        context = DataContext(directory)
        return context
    except (ge_exceptions.ConfigNotFoundError, ge_exceptions.InvalidConfigError) as err:
        cli_message("<red>{}</red>".format(err.message))
        sys.exit(1)
    except ge_exceptions.PluginModuleNotFoundError as err:
        cli_message(err.cli_colored_message)
        sys.exit(1)
    except ge_exceptions.PluginClassNotFoundError as err:
        cli_message(err.cli_colored_message)
        sys.exit(1)



from great_expectations.core.usage_statistics.anonymizers.anonymizer import Anonymizer
from great_expectations.dataset import Dataset

GE_EXPECTATION_TYPES = [
    el for el in Dataset.__dict__.keys() if el.startswith("expect_")
]


class ExpectationSuiteAnonymizer(Anonymizer):
    def __init__(self, salt=None):
        super(ExpectationSuiteAnonymizer, self).__init__(salt=salt)
        self._ge_expectation_types = GE_EXPECTATION_TYPES

    def anonymize_expectation_suite_info(self, expectation_suite):
        anonymized_info_dict = dict()
        anonymized_expectation_counts = dict()

        expectations = expectation_suite.expectations
        expectation_types = [
            expectation.expectation_type for expectation in expectations
        ]
        for expectation_type in set(expectation_types):
            if expectation_type in self._ge_expectation_types:
                anonymized_expectation_counts[expectation_type] = expectation_types.count(expectation_type)
            else:
                anonymized_expectation_type = self.anonymize(expectation_type)
                anonymized_expectation_counts[anonymized_expectation_type] = expectation_types.count(expectation_type)

        anonymized_info_dict["anonymized_name"] = self.anonymize(
            expectation_suite.expectation_suite_name
        )
        anonymized_info_dict["expectation_count"] = len(expectations)
        anonymized_info_dict["anonymized_expectation_type_counts"] = anonymized_expectation_counts

        return anonymized_info_dict




import uuid

from great_expectations.core.usage_statistics.anonymizers.anonymizer import Anonymizer


def test_anonymizer_no_salt():
    # No salt will generate a random one.
    anonymizer1 = Anonymizer()
    anonymizer2 = Anonymizer()

    test_name = "i_am_a_name"

    anon_name_1 = anonymizer1.anonymize(test_name)
    anon_name_2 = anonymizer2.anonymize(test_name)
    assert anon_name_1 != anon_name_2
    assert len(anon_name_1) == 32
    assert len(anon_name_2) == 32

    # Provided different salt will produce different results
    anonymizer1 = Anonymizer("hello, friend")
    anonymizer2 = Anonymizer("hello, enemy")

    test_name = "i_am_a_name"

    anon_name_1 = anonymizer1.anonymize(test_name)
    anon_name_2 = anonymizer2.anonymize(test_name)
    assert anon_name_1 != anon_name_2
    assert len(anon_name_1) == 32
    assert len(anon_name_2) == 32


def test_anonymizer_consistent_salt():
    # Provided same salt will produce same results
    data_context_id = str(uuid.uuid4())
    anonymizer1 = Anonymizer(data_context_id)
    anonymizer2 = Anonymizer(data_context_id)

    test_name = "i_am_a_name"

    anon_name_1 = anonymizer1.anonymize(test_name)
    anon_name_2 = anonymizer2.anonymize(test_name)
    assert anon_name_1 == anon_name_2
    assert len(anon_name_1) == 32
    assert len(anon_name_2) == 32



import logging

from marshmallow import Schema, fields, post_load

from great_expectations.core import IDDict
from great_expectations.core.data_context_key import DataContextKey
from great_expectations.exceptions import InvalidDataContextKeyError, DataContextError

logger = logging.getLogger(__name__)


class ExpectationSuiteIdentifier(DataContextKey):
    def __init__(self, expectation_suite_name):
        super(ExpectationSuiteIdentifier, self).__init__()
        self._expectation_suite_name = expectation_suite_name

    @property
    def expectation_suite_name(self):
        return self._expectation_suite_name

    def to_tuple(self):
        return tuple(self.expectation_suite_name.split("."))

    def to_fixed_length_tuple(self):
        return self.expectation_suite_name,

    @classmethod
    def from_tuple(cls, tuple_):
        return cls(".".join(tuple_))

    @classmethod
    def from_fixed_length_tuple(cls, tuple_):
        return cls(expectation_suite_name=tuple_[0])


class ExpectationSuiteIdentifierSchema(Schema):
    expectation_suite_name = fields.Str()

    # noinspection PyUnusedLocal
    @post_load
    def make_expectation_suite_identifier(self, data, **kwargs):
        return ExpectationSuiteIdentifier(**data)


class BatchIdentifier(DataContextKey):
    def __init__(self, batch_identifier):
        super(BatchIdentifier, self).__init__()
        # batch_kwargs
        # if isinstance(batch_identifier, (BatchKwargs, dict)):
        #     self._batch_identifier = batch_identifier.batch_fingerprint
        # else:
        self._batch_identifier = batch_identifier

    @property
    def batch_identifier(self):
        return self._batch_identifier

    def to_tuple(self):
        return self.batch_identifier,

    @classmethod
    def from_tuple(cls, tuple_):
        return cls(batch_identifier=tuple_[0])


class BatchIdentifierSchema(Schema):
    batch_identifier = fields.Str()

    # noinspection PyUnusedLocal
    @post_load
    def make_batch_identifier(self, data, **kwargs):
        return BatchIdentifier(**data)


class ValidationResultIdentifier(DataContextKey):
    """A ValidationResultIdentifier identifies a validation result by the fully qualified expectation_suite_identifer
    and run_id.
    """

    def __init__(self, expectation_suite_identifier, run_id, batch_identifier):
        """Constructs a ValidationResultIdentifier

        Args:
            expectation_suite_identifier (ExpectationSuiteIdentifier, list, tuple, or dict):
                identifying information for the fully qualified expectation suite used to validate
            run_id (str): The run_id for which validation occurred
        """
        super(ValidationResultIdentifier, self).__init__()
        self._expectation_suite_identifier = expectation_suite_identifier
        self._run_id = run_id
        self._batch_identifier = batch_identifier

    @property
    def expectation_suite_identifier(self):
        return self._expectation_suite_identifier

    @property
    def run_id(self):
        return self._run_id

    @property
    def batch_identifier(self):
        return self._batch_identifier

    def to_tuple(self):
        return tuple(
            list(self.expectation_suite_identifier.to_tuple()) + [
                self.run_id or "__none__",
                self.batch_identifier or "__none__"
            ]
        )

    def to_fixed_length_tuple(self):
        return self.expectation_suite_identifier.expectation_suite_name, self.run_id or "__none__", \
               self.batch_identifier or "__none__"

    @classmethod
    def from_tuple(cls, tuple_):
        return cls(ExpectationSuiteIdentifier.from_tuple(tuple_[0:-2]), tuple_[-2], tuple_[-1])

    @classmethod
    def from_fixed_length_tuple(cls, tuple_):
        return cls(ExpectationSuiteIdentifier(tuple_[0]), tuple_[1], tuple_[2])

    @classmethod
    def from_object(cls, validation_result):
        batch_kwargs = validation_result.meta.get("batch_kwargs", {})
        if isinstance(batch_kwargs, IDDict):
            batch_identifier = batch_kwargs.to_id()
        elif isinstance(batch_kwargs, dict):
            batch_identifier = IDDict(batch_kwargs).to_id()
        else:
            raise DataContextError("Unable to construct ValidationResultIdentifier from provided object.")
        return cls(
            expectation_suite_identifier=ExpectationSuiteIdentifier(validation_result.meta["expectation_suite_name"]),
            run_id=validation_result.meta.get("run_id"),
            batch_identifier=batch_identifier
        )


class ValidationResultIdentifierSchema(Schema):
    expectation_suite_identifier = fields.Nested(ExpectationSuiteIdentifierSchema, required=True, error_messages={
        'required': 'expectation_suite_identifier is required for a ValidationResultIdentifier'})
    run_id = fields.Str(required=True, error_messages={'required': "run_id is required for a "
                                                                   "ValidationResultIdentifier"})
    batch_identifier = fields.Nested(BatchIdentifierSchema, required=True)

    # noinspection PyUnusedLocal
    @post_load
    def make_validation_result_identifier(self, data, **kwargs):
        return ValidationResultIdentifier(**data)


class SiteSectionIdentifier(DataContextKey):
    def __init__(self, site_section_name, resource_identifier):
        self._site_section_name = site_section_name
        if site_section_name in ["validations", "profiling"]:
            if isinstance(resource_identifier, ValidationResultIdentifier):
                self._resource_identifier = resource_identifier
            elif isinstance(resource_identifier, (tuple, list)):
                self._resource_identifier = ValidationResultIdentifier(*resource_identifier)
            else:
                self._resource_identifier = ValidationResultIdentifier(**resource_identifier)
        elif site_section_name == "expectations":
            if isinstance(resource_identifier, ExpectationSuiteIdentifier):
                self._resource_identifier = resource_identifier
            elif isinstance(resource_identifier, (tuple, list)):
                self._resource_identifier = ExpectationSuiteIdentifier(*resource_identifier)
            else:
                self._resource_identifier = ExpectationSuiteIdentifier(**resource_identifier)
        else:
            raise InvalidDataContextKeyError(
                "SiteSectionIdentifier only supports 'validations' and 'expectations' as site section names"
            )

    @property
    def site_section_name(self):
        return self._site_section_name

    @property
    def resource_identifier(self):
        return self._resource_identifier

    def to_tuple(self):
        site_section_identifier_tuple_list = [self.site_section_name] + list(self.resource_identifier.to_tuple())
        return tuple(site_section_identifier_tuple_list)

    @classmethod
    def from_tuple(cls, tuple_):
        if tuple_[0] == "validations":
            return cls(
                site_section_name=tuple_[0],
                resource_identifier=ValidationResultIdentifier.from_tuple(tuple_[1:])
            )
        elif tuple_[0] == "expectations":
            return cls(
                site_section_name=tuple_[0],
                resource_identifier=ExpectationSuiteIdentifier.from_tuple(tuple_[1:])
            )
        else:
            raise InvalidDataContextKeyError(
                "SiteSectionIdentifier only supports 'validations' and 'expectations' as site section names"
            )


expectationSuiteIdentifierSchema = ExpectationSuiteIdentifierSchema()
validationResultIdentifierSchema = ValidationResultIdentifierSchema()



import pytest

from great_expectations import __version__ as ge_version
from great_expectations.core import ExpectationSuite, ExpectationConfiguration, ExpectationKwargs
from great_expectations.data_asset import DataAsset, FileDataAsset
from great_expectations.dataset import Dataset, PandasDataset


def test_data_asset_expectation_suite():
    asset = DataAsset()
    default_suite = ExpectationSuite(
        expectation_suite_name="default",
        data_asset_type="DataAsset",
        meta={
            "great_expectations.__version__": ge_version
        },
        expectations=[]
    )

    # We should have a default-initialized suite stored internally and available for getting
    assert asset._expectation_suite == default_suite
    assert asset.get_expectation_suite() == default_suite


def test_interactive_evaluation(dataset):
    # We should be able to enable and disable interactive evaluation

    # Default is on
    assert dataset.get_config_value("interactive_evaluation") is True
    res = dataset.expect_column_values_to_be_between("naturals", 1, 10, include_config=True)
    assert res.success is True

    # Disable
    dataset.set_config_value("interactive_evaluation", False)
    disable_res = dataset.expect_column_values_to_be_between("naturals", 1, 10)  # No need to explicitly include_config
    assert disable_res.success is None

    assert res.expectation_config == disable_res.expectation_config


def test_data_asset_name_inheritance(dataset):
    # A data_asset should have a generic type
    data_asset = DataAsset()
    assert data_asset.get_expectation_suite().data_asset_type == "DataAsset"

    # A FileDataAsset should pick up its type
    data_asset = FileDataAsset()
    assert data_asset.get_expectation_suite().data_asset_type == "FileDataAsset"

    # So should a Dataset
    data_asset = Dataset()
    assert data_asset.get_expectation_suite().data_asset_type == "Dataset"

    # Backends should *not* change the implementation
    assert dataset.get_expectation_suite().data_asset_type == "Dataset"

    # But custom classes should choose to
    class MyCustomDataset(Dataset):
        _data_asset_type = "MyCustomDataset"

    data_asset = MyCustomDataset()
    assert data_asset.get_expectation_suite().data_asset_type == "MyCustomDataset"


def test_catch_exceptions_with_bad_expectation_type():
    # We want to catch degenerate cases where an expectation suite is incompatible with
    my_df = PandasDataset({"x": range(10)})
    my_df._append_expectation(ExpectationConfiguration(expectation_type='foobar', kwargs={}))
    result = my_df.validate(catch_exceptions=True)

    # Find the foobar result
    idx = 0
    for idx, val_result in enumerate(result.results):
        if val_result.expectation_config.expectation_type == "foobar":
            break

    assert result.results[idx].success is False
    assert result.results[idx].expectation_config.expectation_type == "foobar"
    assert result.results[idx].expectation_config.kwargs == ExpectationKwargs()
    assert result.results[idx].exception_info["raised_exception"] is True
    assert "AttributeError: \'PandasDataset\' object has no attribute \'foobar\'" in \
           result.results[idx].exception_info["exception_traceback"]

    with pytest.raises(AttributeError):
        result = my_df.validate(catch_exceptions=False)


def test_valid_expectation_types(dataset, pandas_dataset):
    assert pandas_dataset.list_available_expectation_types() == [
        "expect_column_bootstrapped_ks_test_p_value_to_be_greater_than",
        "expect_column_chisquare_test_p_value_to_be_greater_than",
        "expect_column_distinct_values_to_be_in_set",
        "expect_column_distinct_values_to_contain_set",
        "expect_column_distinct_values_to_equal_set",
        "expect_column_kl_divergence_to_be_less_than",
        "expect_column_max_to_be_between",
        "expect_column_mean_to_be_between",
        "expect_column_median_to_be_between",
        "expect_column_min_to_be_between",
        "expect_column_most_common_value_to_be_in_set",
        "expect_column_pair_values_A_to_be_greater_than_B",
        "expect_column_pair_values_to_be_equal",
        "expect_column_pair_values_to_be_in_set",
        "expect_column_parameterized_distribution_ks_test_p_value_to_be_greater_than",
        "expect_column_proportion_of_unique_values_to_be_between",
        "expect_column_quantile_values_to_be_between",
        "expect_column_stdev_to_be_between",
        "expect_column_sum_to_be_between",
        "expect_column_to_exist",
        "expect_column_unique_value_count_to_be_between",
        "expect_column_value_lengths_to_be_between",
        "expect_column_value_lengths_to_equal",
        "expect_column_values_to_be_between",
        "expect_column_values_to_be_dateutil_parseable",
        "expect_column_values_to_be_decreasing",
        "expect_column_values_to_be_in_set",
        "expect_column_values_to_be_in_type_list",
        "expect_column_values_to_be_increasing",
        "expect_column_values_to_be_json_parseable",
        "expect_column_values_to_be_null",
        "expect_column_values_to_be_of_type",
        "expect_column_values_to_be_unique",
        "expect_column_values_to_match_json_schema",
        "expect_column_values_to_match_regex",
        "expect_column_values_to_match_regex_list",
        "expect_column_values_to_match_strftime_format",
        "expect_column_values_to_not_be_in_set",
        "expect_column_values_to_not_be_null",
        "expect_column_values_to_not_match_regex",
        "expect_column_values_to_not_match_regex_list",
        "expect_multicolumn_values_to_be_unique",
        "expect_table_column_count_to_be_between",
        "expect_table_column_count_to_equal",
        "expect_table_columns_to_match_ordered_list",
        "expect_table_row_count_to_be_between",
        "expect_table_row_count_to_equal",
    ]



import os
from typing import Union

import nbformat

from great_expectations.core import ExpectationSuite
from great_expectations.core.id_dict import BatchKwargs
from great_expectations.render.renderer.renderer import Renderer
from great_expectations.util import lint_code


class SuiteEditNotebookRenderer(Renderer):
    """
    Render a notebook that can re-create or edit a suite.

    Use cases:
    - Make an easy path to edit a suite that a Profiler created.
    - Make it easy to edit a suite where only JSON exists.
    """

    @classmethod
    def _get_expectations_by_column(cls, expectations):
        # TODO probably replace this with Suite logic at some point
        expectations_by_column = {"table_expectations": []}
        for exp in expectations:
            if "column" in exp["kwargs"]:
                col = exp["kwargs"]["column"]

                if col not in expectations_by_column.keys():
                    expectations_by_column[col] = []
                expectations_by_column[col].append(exp)
            else:
                expectations_by_column["table_expectations"].append(exp)

        return expectations_by_column

    @classmethod
    def _build_kwargs_string(cls, expectation):
        kwargs = []
        for k, v in expectation["kwargs"].items():
            if k == "column":
                # make the column a positional argument
                kwargs.append("'{}'".format(v))
            elif isinstance(v, str):
                # Put strings in quotes
                kwargs.append("{}='{}'".format(k, v))
            else:
                # Pass other types as is
                kwargs.append("{}={}".format(k, v))

        return ", ".join(kwargs)

    def add_header(self, suite_name: str, batch_kwargs) -> None:
        self.add_markdown_cell(
            """# Edit Your Expectation Suite
Use this notebook to recreate and modify your expectation suite:

**Expectation Suite Name**: `{}`

We'd love it if you **reach out to us on** the [**Great Expectations Slack Channel**](https://greatexpectations.io/slack)""".format(
                suite_name
            )
        )

        if not batch_kwargs:
            batch_kwargs = dict()
        self.add_code_cell(
            """\
from datetime import datetime
import great_expectations as ge
import great_expectations.jupyter_ux
from great_expectations.data_context.types.resource_identifiers import ValidationResultIdentifier

context = ge.data_context.DataContext()

# Feel free to change the name of your suite here. Renaming this will not
# remove the other one.
expectation_suite_name = "{}"
suite = context.get_expectation_suite(expectation_suite_name)
suite.expectations = []

batch_kwargs = {}
batch = context.get_batch(batch_kwargs, suite)
batch.head()""".format(
                suite_name, batch_kwargs
            ),
            lint=True,
        )

    def add_footer(self) -> None:
        self.add_markdown_cell(
            """\
## Save & Review Your Expectations

Let's save the expectation suite as a JSON file in the `great_expectations/expectations` directory of your project.
If you decide not to save some expectations that you created, use [remove_expectation method](https://docs.greatexpectations.io/en/latest/module_docs/data_asset_module.html?highlight=remove_expectation&utm_source=notebook&utm_medium=edit_expectations#great_expectations.data_asset.data_asset.DataAsset.remove_expectation).

Let's now rebuild your Data Docs, which helps you communicate about your data with both machines and humans."""
        )
        # TODO this may become confusing for users depending on what they are trying
        #  to accomplish in their dev loop
        self.add_code_cell(
            """\
batch.save_expectation_suite(discard_failed_expectations=False)

# Let's make a simple sortable timestamp. Note this could come from your pipeline runner.
run_id = datetime.utcnow().strftime("%Y%m%dT%H%M%S.%fZ")

results = context.run_validation_operator("action_list_operator", assets_to_validate=[batch], run_id=run_id)
expectation_suite_identifier = list(results["details"].keys())[0]
validation_result_identifier = ValidationResultIdentifier(
    expectation_suite_identifier=expectation_suite_identifier,
    batch_identifier=batch.batch_kwargs.to_id(),
    run_id=run_id
)
context.build_data_docs()
context.open_data_docs(validation_result_identifier)"""
        )

    def add_code_cell(self, code: str, lint: bool = False) -> None:
        """
        Add the given code as a new code cell.
        """
        if lint:
            code = lint_code(code).rstrip("\n")

        cell = nbformat.v4.new_code_cell(code)
        self._notebook["cells"].append(cell)

    def add_markdown_cell(self, markdown: str) -> None:
        """
        Add the given markdown as a new markdown cell.
        """
        cell = nbformat.v4.new_markdown_cell(markdown)
        self._notebook["cells"].append(cell)

    def add_expectation_cells_from_suite(self, expectations):
        expectations_by_column = self._get_expectations_by_column(expectations)
        self.add_markdown_cell("### Table Expectation(s)")
        self._add_table_level_expectations(expectations_by_column)
        # Remove the table expectations since they are dealt with
        expectations_by_column.pop("table_expectations")
        self.add_markdown_cell("### Column Expectation(s)")
        self._add_column_level_expectations(expectations_by_column)

    def _add_column_level_expectations(self, expectations_by_column):
        if not expectations_by_column:
            self.add_markdown_cell(
                "No column level expectations are in this suite. Feel free to "
                "add some here. They all begin with `batch.expect_column_...`."
            )
            return

        for column, expectations in expectations_by_column.items():
            self.add_markdown_cell("#### `{}`".format(column))

            for exp in expectations:
                kwargs_string = self._build_kwargs_string(exp)
                meta_args = self._build_meta_arguments(exp.meta)
                code = "batch.{}({}{})".format(
                    exp["expectation_type"], kwargs_string, meta_args
                )
                self.add_code_cell(code, lint=True)

    def _add_table_level_expectations(self, expectations_by_column):
        if not expectations_by_column["table_expectations"]:
            self.add_markdown_cell(
                "No table level expectations are in this suite. Feel free to "
                "add some here. They all begin with `batch.expect_table_...`."
            )
            return

        for exp in expectations_by_column["table_expectations"]:
            kwargs_string = self._build_kwargs_string(exp)
            code = "batch.{}({})".format(exp["expectation_type"], kwargs_string)
            self.add_code_cell(code, lint=True)

    @staticmethod
    def _build_meta_arguments(meta):
        if not meta:
            return ""

        profiler = "BasicSuiteBuilderProfiler"
        if profiler in meta.keys():
            meta.pop(profiler)

        if meta.keys():
            return ", meta={}".format(meta)

        return ""

    @classmethod
    def write_notebook_to_disk(cls, notebook, notebook_file_path):
        with open(notebook_file_path, "w") as f:
            nbformat.write(notebook, f)

    def render(
        self, suite: ExpectationSuite, batch_kwargs=None
    ) -> nbformat.NotebookNode:
        """
        Render a notebook dict from an expectation suite.
        """
        if not isinstance(suite, ExpectationSuite):
            raise RuntimeWarning("render must be given an ExpectationSuite.")

        self._notebook = nbformat.v4.new_notebook()

        suite_name = suite.expectation_suite_name

        batch_kwargs = self.get_batch_kwargs(suite, batch_kwargs)
        self.add_header(suite_name, batch_kwargs)
        self.add_authoring_intro()
        self.add_expectation_cells_from_suite(suite.expectations)
        self.add_footer()

        return self._notebook

    def render_to_disk(
        self, suite: ExpectationSuite, notebook_file_path: str, batch_kwargs=None
    ) -> None:
        """
        Render a notebook to disk from an expectation suite.

        If batch_kwargs are passed they will override any found in suite
        citations.
        """
        self.render(suite, batch_kwargs)
        self.write_notebook_to_disk(self._notebook, notebook_file_path)

    def add_authoring_intro(self):
        self.add_markdown_cell(
            """\
## Create & Edit Expectations

Add expectations by calling specific expectation methods on the `batch` object. They all begin with `.expect_` which makes autocompleting easy using tab.

You can see all the available expectations in the **[expectation glossary](https://docs.greatexpectations.io/en/latest/expectation_glossary.html?utm_source=notebook&utm_medium=create_expectations)**."""
        )

    def get_batch_kwargs(self, suite: ExpectationSuite, batch_kwargs: Union[dict, BatchKwargs]):
        if isinstance(batch_kwargs, dict):
            return self._fix_path_in_batch_kwargs(batch_kwargs)

        citations = suite.meta.get("citations")
        if not citations:
            return self._fix_path_in_batch_kwargs(batch_kwargs)

        citations = suite.get_citations(require_batch_kwargs=True)
        if not citations:
            return None

        citation = citations[-1]
        batch_kwargs = citation.get("batch_kwargs")
        return self._fix_path_in_batch_kwargs(batch_kwargs)

    @staticmethod
    def _fix_path_in_batch_kwargs(batch_kwargs):
        if isinstance(batch_kwargs, BatchKwargs):
            batch_kwargs = dict(batch_kwargs)
        if batch_kwargs and "path" in batch_kwargs.keys():
            base_dir = batch_kwargs["path"]
            if not os.path.isabs(base_dir):
                batch_kwargs["path"] = os.path.join("..", "..", base_dir)

        return batch_kwargs



import logging
import urllib
from collections import OrderedDict

import traceback

from great_expectations.render.renderer.call_to_action_renderer import CallToActionRenderer
from great_expectations.render.renderer.renderer import Renderer
from great_expectations.render.types import (
    RenderedSectionContent,
    RenderedDocumentContent,
    RenderedHeaderContent,
    RenderedStringTemplateContent,
    RenderedTableContent,
    RenderedBulletListContent,
)

logger = logging.getLogger(__name__)


# FIXME : This class needs to be rebuilt to accept SiteSectionIdentifiers as input.
# FIXME : This class needs tests.
class SiteIndexPageRenderer(Renderer):

    @classmethod
    def _generate_links_table_rows(cls, index_links_dict, link_list_keys_to_render):
        section_rows = []

        column_count = len(link_list_keys_to_render)
        validations_links = index_links_dict.get("validations_links")
        expectations_links = index_links_dict.get("expectations_links")

        if column_count:
            cell_width_pct = 100.0/column_count

        if "expectations_links" in link_list_keys_to_render:
            for expectation_suite_link_dict in expectations_links:
                expectation_suite_row = []
                expectation_suite_name = expectation_suite_link_dict["expectation_suite_name"]

                expectation_suite_link = RenderedStringTemplateContent(**{
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": "$link_text",
                        "params": {
                            "link_text": expectation_suite_name
                        },
                        "tag": "a",
                        "styling": {
                            "attributes": {
                                "href": urllib.parse.quote(expectation_suite_link_dict["filepath"])
                            },
                            "classes": ["ge-index-page-table-expectation-suite-link"]
                        }
                    },
                    "styling": {
                        "parent": {
                            "styles": {
                                "width": "{}%".format(cell_width_pct),
                            }
                        }
                    }
                })
                expectation_suite_row.append(expectation_suite_link)

                if "validations_links" in link_list_keys_to_render:
                    sorted_validations_links = [
                        link_dict for link_dict in sorted(validations_links, key=lambda x: x["run_id"], reverse=True)
                        if link_dict["expectation_suite_name"] == expectation_suite_name
                    ]
                    validation_link_bullets = [
                        RenderedStringTemplateContent(**{
                            "content_block_type": "string_template",
                            "string_template": {
                                "template": "${validation_success} $link_text",
                                "params": {
                                    "link_text": link_dict["run_id"],
                                    "validation_success": ""
                                },
                                "tag": "a",
                                "styling": {
                                    "attributes": {
                                        "href": urllib.parse.quote(link_dict["filepath"])
                                    },
                                    "params": {
                                        "validation_success": {
                                            "tag": "i",
                                            "classes": ["fas", "fa-check-circle",  "text-success"] if link_dict["validation_success"] else ["fas", "fa-times", "text-danger"]
                                        }
                                    },
                                    "classes": ["ge-index-page-table-validation-links-item"]
                                }
                            },
                            "styling": {
                                "parent": {
                                    "classes": ["hide-succeeded-validation-target"] if link_dict[
                                        "validation_success"] else []
                                }
                            }
                        }) for link_dict in sorted_validations_links if
                        link_dict["expectation_suite_name"] == expectation_suite_name
                    ]
                    validation_link_bullet_list = RenderedBulletListContent(**{
                        "content_block_type": "bullet_list",
                        "bullet_list": validation_link_bullets,
                        "styling": {
                            "parent": {
                                "styles": {
                                    "width": "{}%".format(cell_width_pct)
                                }
                            },
                            "body": {
                                "classes": ["ge-index-page-table-validation-links-list"]
                            }
                        }
                    })
                    expectation_suite_row.append(validation_link_bullet_list)

                section_rows.append(expectation_suite_row)

        if not expectations_links and "validations_links" in link_list_keys_to_render:
            sorted_validations_links = [
                link_dict for link_dict in sorted(validations_links, key=lambda x: x["run_id"], reverse=True)
            ]
            validation_link_bullets = [
                RenderedStringTemplateContent(**{
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": "${validation_success} $link_text",
                        "params": {
                            "link_text": link_dict["run_id"],
                            "validation_success": ""
                        },
                        "tag": "a",
                        "styling": {
                            "attributes": {
                                "href": urllib.parse.quote(link_dict["filepath"])
                            },
                            "params": {
                                "validation_success": {
                                    "tag": "i",
                                    "classes": ["fas", "fa-check-circle", "text-success"] if link_dict[
                                        "validation_success"] else ["fas", "fa-times", "text-danger"]
                                }
                            },
                            "classes": ["ge-index-page-table-validation-links-item"]
                        }
                    },
                    "styling": {
                        "parent": {
                            "classes": ["hide-succeeded-validation-target"] if link_dict[
                                "validation_success"] else []
                        }
                    }
                }) for link_dict in sorted_validations_links
            ]
            validation_link_bullet_list = RenderedBulletListContent(**{
                "content_block_type": "bullet_list",
                "bullet_list": validation_link_bullets,
                "styling": {
                    "parent": {
                        "styles": {
                            "width": "{}%".format(cell_width_pct)
                        }
                    },
                    "body": {
                        "classes": ["ge-index-page-table-validation-links-list"]
                    }
                }
            })
            section_rows.append([validation_link_bullet_list])

        return section_rows

    @classmethod
    def render(cls, index_links_dict):
        sections = []
        cta_object = index_links_dict.pop("cta_object", None)

        try:
            content_blocks = []
            # site name header
            site_name_header_block = RenderedHeaderContent(**{
                "content_block_type": "header",
                "header": RenderedStringTemplateContent(**{
                    "content_block_type": "string_template",
                    "string_template": {
                        "template": "$title_prefix | $site_name",
                        "params": {
                            "site_name": index_links_dict.get("site_name"),
                            "title_prefix": "Data Docs"
                        },
                        "styling": {
                            "params": {
                                "title_prefix": {
                                    "tag": "strong"
                                }
                            }
                        },
                    }
                }),
                "styling": {
                    "classes": ["col-12", "ge-index-page-site-name-title"],
                    "header": {
                        "classes": ["alert", "alert-secondary"]
                    }
                }
            })
            content_blocks.append(site_name_header_block)

            table_rows = []
            table_header_row = []
            link_list_keys_to_render = []

            header_dict = OrderedDict([
                ["expectations_links", "Expectation Suite"],
                ["validations_links", "Validation Results (run_id)"]
            ])

            for link_lists_key, header in header_dict.items():
                if index_links_dict.get(link_lists_key):
                    class_header_str = link_lists_key.replace("_", "-")
                    class_str = "ge-index-page-table-{}-header".format(class_header_str)
                    header = RenderedStringTemplateContent(**{
                        "content_block_type": "string_template",
                        "string_template": {
                            "template": header,
                            "params": {},
                            "styling": {
                                "classes": [class_str],
                            }
                        }
                    })
                    table_header_row.append(header)
                    link_list_keys_to_render.append(link_lists_key)

            generator_table = RenderedTableContent(**{
                "content_block_type": "table",
                "header_row": table_header_row,
                "table": table_rows,
                "styling": {
                    "classes": ["col-12", "ge-index-page-table-container"],
                    "styles": {
                        "margin-top": "10px"
                    },
                    "body": {
                        "classes": ["table", "table-sm", "ge-index-page-generator-table"]
                    }
                }
            })

            table_rows += cls._generate_links_table_rows(
                index_links_dict, link_list_keys_to_render=link_list_keys_to_render
            )

            content_blocks.append(generator_table)

            if index_links_dict.get("profiling_links"):
                profiling_table_rows = []
                for profiling_link_dict in index_links_dict.get("profiling_links"):
                    profiling_table_rows.append(
                        [
                            RenderedStringTemplateContent(**{
                                "content_block_type": "string_template",
                                "string_template": {
                                    "template": "$link_text",
                                    "params": {
                                        "link_text": profiling_link_dict["expectation_suite_name"] + "." + profiling_link_dict["batch_identifier"]
                                    },
                                    "tag": "a",
                                    "styling": {
                                        "attributes": {
                                            "href": urllib.parse.quote(profiling_link_dict["filepath"])
                                        },
                                        "classes": ["ge-index-page-table-expectation-suite-link"]
                                    }
                                },
                            })
                        ]
                    )
                content_blocks.append(
                    RenderedTableContent(**{
                        "content_block_type": "table",
                        "header_row": ["Profiling Results"],
                        "table": profiling_table_rows,
                        "styling": {
                            "classes": ["col-12", "ge-index-page-table-container"],
                            "styles": {
                                "margin-top": "10px"
                            },
                            "body": {
                                "classes": ["table", "table-sm", "ge-index-page-generator-table"]
                            }
                        }
                    })

                )

            section = RenderedSectionContent(**{
                "section_name": index_links_dict.get("site_name"),
                "content_blocks": content_blocks
            })
            sections.append(section)

            index_page_document = RenderedDocumentContent(**{
                "renderer_type": "SiteIndexPageRenderer",
                "utm_medium": "index-page",
                "sections": sections
            })

            if cta_object:
                index_page_document.cta_footer = CallToActionRenderer.render(cta_object)

            return index_page_document

        except Exception as e:
            exception_message = f'''\
An unexpected Exception occurred during data docs rendering.  Because of this error, certain parts of data docs will \
not be rendered properly and/or may not appear altogether.  Please use the trace, included in this message, to \
diagnose and repair the underlying issue.  Detailed information follows:  
            '''
            exception_traceback = traceback.format_exc()
            exception_message += f'{type(e).__name__}: "{str(e)}".  Traceback: "{exception_traceback}".'
            logger.error(exception_message, e, exc_info=True)



import copy
import datetime
import os
import signal
import subprocess
import time
import uuid

import boto3
import botocore
import pytest
import requests

from great_expectations.data_context.util import file_relative_path

USAGE_STATISTICS_QA_URL = "https://qa.stats.greatexpectations.io/great_expectations/v1/usage_statistics"

logGroupName = "/great_expectations/usage_statistics/qa"


@pytest.fixture(scope="session")
def aws_session():
    aws_session = None
    for session_options in [
        {"profile_name": "travis-ci", "region_name": "us-east-1"},
        {"profile_name": "default", "region_name": "us-east-1"},
        {"region_name": "us-east-1"},
    ]:
        try:
            aws_session = boto3.Session(**session_options)
        except botocore.exceptions.ProfileNotFound:
            continue
    return aws_session


@pytest.fixture(scope="session")
def valid_usage_statistics_message():
    return {
        'event_payload': {
            'platform.system': 'Darwin',
            'platform.release': '19.3.0',
            'version_info': "sys.version_info(major=3, minor=7, micro=4, releaselevel='final', serial=0)",
            'anonymized_datasources': [
            {
                'anonymized_name': 'f57d8a6edae4f321b833384801847498',
                'parent_class': 'SqlAlchemyDatasource',
                'sqlalchemy_dialect': 'postgresql'
            }
        ],
        'anonymized_stores': [
            {
                'anonymized_name': '078eceafc1051edf98ae2f911484c7f7',
                'parent_class': 'ExpectationsStore',
                'anonymized_store_backend': {
                    'parent_class': 'TupleFilesystemStoreBackend'
                }
            },
            {
                'anonymized_name': '313cbd9858dd92f3fc2ef1c10ab9c7c8',
                'parent_class': 'ValidationsStore',
                'anonymized_store_backend': {
                    'parent_class': 'TupleFilesystemStoreBackend'
                }
            },
            {
                'anonymized_name': '2d487386aa7b39e00ed672739421473f',
                'parent_class': 'EvaluationParameterStore',
                'anonymized_store_backend': {
                    'parent_class': 'InMemoryStoreBackend'
                }
            }
        ],
        'anonymized_validation_operators': [
            {
                'anonymized_name': '99d14cc00b69317551690fb8a61aca94',
                'parent_class': 'ActionListValidationOperator',
                'anonymized_action_list': [
                    {
                        'anonymized_name': '5a170e5b77c092cc6c9f5cf2b639459a',
                        'parent_class': 'StoreValidationResultAction'
                    },
                    {
                        'anonymized_name': '0fffe1906a8f2a5625a5659a848c25a3',
                        'parent_class': 'StoreEvaluationParametersAction'
                    },
                    {
                        'anonymized_name': '101c746ab7597e22b94d6e5f10b75916',
                        'parent_class': 'UpdateDataDocsAction'
                    }
                ]
            }
        ],
        'anonymized_data_docs_sites': [
            {
                'parent_class': 'SiteBuilder',
                'anonymized_name': 'eaf0cf17ad63abf1477f7c37ad192700',
                'anonymized_store_backend': {'parent_class': 'TupleFilesystemStoreBackend'},
                'anonymized_site_index_builder': {
                    'parent_class': 'DefaultSiteIndexBuilder',
                    'show_cta_footer': True
                }
            }
        ],
        'anonymized_expectation_suites': [
            {
                'anonymized_name': '238e99998c7674e4ff26a9c529d43da4',
                'expectation_count': 8,
                'anonymized_expectation_type_counts': {
                    'expect_column_value_lengths_to_be_between': 1,
                    'expect_table_row_count_to_be_between': 1,
                    'expect_column_values_to_not_be_null': 2,
                    'expect_column_distinct_values_to_be_in_set': 1,
                    'expect_column_kl_divergence_to_be_less_than': 1,
                    'expect_table_column_count_to_equal': 1,
                    'expect_table_columns_to_match_ordered_list': 1
                }
            }
        ]
    },
    'event': 'data_context.__init__',
    'success': True,
    'version': '1.0.0',
    'event_time': '2020-03-28T01:14:21.155Z',
    'data_context_id': '96c547fe-e809-4f2e-b122-0dc91bb7b3ad',
    'data_context_instance_id': '445a8ad1-2bd0-45ce-bb6b-d066afe996dd',
    'ge_version': '0.9.7+244.g56d67e51d.dirty'
}


def test_send_malformed_data(valid_usage_statistics_message):
    # We should be able to successfully send a valid message, but find that
    # a malformed message is not accepted
    res = requests.post(USAGE_STATISTICS_QA_URL, json=valid_usage_statistics_message)
    assert res.status_code == 201
    invalid_usage_statistics_message = copy.deepcopy(valid_usage_statistics_message)
    del invalid_usage_statistics_message["data_context_id"]
    res = requests.post(USAGE_STATISTICS_QA_URL, json=invalid_usage_statistics_message)
    assert res.status_code == 400


@pytest.mark.aws_integration
def test_usage_statistics_transmission(aws_session):
    client = aws_session.client('logs', region_name='us-east-1')
    usage_stats_url_env = dict(**os.environ)
    usage_stats_url_env["GE_USAGE_STATISTICS_URL"] = USAGE_STATISTICS_QA_URL
    data_context_id = str(uuid.uuid4())
    p = subprocess.Popen(
        ["python", file_relative_path(__file__, "./instantiate_context_with_usage_statistics.py"), data_context_id,
         "0"],
        stdout=subprocess.PIPE, stderr=subprocess.PIPE,
        env=usage_stats_url_env
    )
    outs, errs = p.communicate()
    outs = str(outs)
    errs = str(errs)
    assert "INFO" not in outs
    assert "Done constructing a DataContext" in outs
    assert "Ending a long nap" in outs
    assert "KeyboardInterrupt" not in errs
    assert errs.count("201") == 4
    print(data_context_id)
    time.sleep(60)
    queryId = client.start_query(
        logGroupName=logGroupName,
        startTime=int((datetime.datetime.now() - datetime.timedelta(minutes=10)).timestamp()),
        endTime=int((datetime.datetime.now() - datetime.timedelta(seconds=1)).timestamp()),
        queryString='fields @timestamp, @message | filter data_context_id = "' + data_context_id + '"',
        limit=40
    ).get("queryId")
    done = False
    tries = 0
    while tries < 3 and not done:
        time.sleep(5)
        response = client.get_query_results(
            queryId=queryId
        )
        tries += 1
        done = response["status"] == "Complete"

    print(response)
    assert response["statistics"]["recordsMatched"] == 4


@pytest.mark.aws_integration
def test_send_completes_on_kill(aws_session):
    client = aws_session.client('logs', region_name='us-east-1')
    """Test that having usage statistics enabled does not negatively impact kill signals or cause loss of queued usage statistics. """
    # Execute process that initializes data context
    acceptable_startup_time = 6
    acceptable_shutdown_time = 1
    nap_time = 30
    start = datetime.datetime.now()
    usage_stats_url_env = dict(**os.environ)
    usage_stats_url_env["GE_USAGE_STATISTICS_URL"] = USAGE_STATISTICS_QA_URL
    data_context_id = str(uuid.uuid4())
    # Instruct the process to wait for 30 seconds after initializing before completing.
    p = subprocess.Popen(
        ["python", file_relative_path(__file__, "./instantiate_context_with_usage_statistics.py"),
         data_context_id, str(nap_time), "False", "True"],
        stdout=subprocess.PIPE, stderr=subprocess.PIPE,
        env=usage_stats_url_env
    )
    time.sleep(acceptable_startup_time)
    # Send a signal to kill
    p.send_signal(signal.SIGINT)
    outs, errs = p.communicate()
    end = datetime.datetime.now()

    # Ensure that the process shut down earlier than it would have
    assert datetime.timedelta(
        seconds=acceptable_startup_time
    ) < (end - start) < datetime.timedelta(
        seconds=acceptable_startup_time + acceptable_shutdown_time)

    outs = str(outs)
    errs = str(errs)
    assert "INFO" not in outs
    assert "Done constructing a DataContext" in outs
    assert "Ending a long nap" not in outs
    assert "KeyboardInterrupt" in errs
    assert errs.count("201") == 4
    # An estimate for how long indexing takes...
    time.sleep(60)
    print(data_context_id)
    queryId = client.start_query(
        logGroupName=logGroupName,
        startTime=int((datetime.datetime.now() - datetime.timedelta(minutes=10)).timestamp()),
        endTime=int((datetime.datetime.now() - datetime.timedelta(seconds=1)).timestamp()),
        queryString='fields @timestamp, @message | filter data_context_id = "' + data_context_id + '"',
        limit=40
    ).get("queryId")
    done = False
    tries = 0
    while tries < 3 and not done:
        time.sleep(5)
        response = client.get_query_results(
            queryId=queryId
        )
        tries += 1
        done = response["status"] == "Complete"

    print(response)
    assert response["statistics"]["recordsMatched"] == 4


def test_graceful_failure_with_no_internet():
    """Test that having usage statistics enabled does not negatively impact kill signals or cause loss of queued usage statistics. """

    # Execute process that initializes data context
    # NOTE - JPC - 20200227 - this is crazy long (not because of logging I think, but worth revisiting)
    acceptable_startup_time = 6
    acceptable_shutdown_time = 1
    nap_time = 0
    start = datetime.datetime.now()
    data_context_id = str(uuid.uuid4())
    # Instruct the process to wait for 30 seconds after initializing before completing.
    p = subprocess.Popen(
        ["python", file_relative_path(__file__, "./instantiate_context_with_usage_statistics.py"),
         data_context_id, str(nap_time), "True", "True"],
        stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    outs, errs = p.communicate()
    end = datetime.datetime.now()
    # We didn't wait or send a signal, so just check that times were reasonable
    assert (end - start) < datetime.timedelta(
        seconds=acceptable_startup_time + acceptable_shutdown_time)
    outs = str(outs)
    errs = str(errs)
    assert "INFO" not in outs
    assert "Done constructing a DataContext" in outs
    assert "Ending a long nap" in outs



import re
import datetime
import logging

try:
    import boto3
except ImportError:
    boto3 = None

from great_expectations.exceptions import GreatExpectationsError
from great_expectations.datasource.batch_kwargs_generator.batch_kwargs_generator import BatchKwargsGenerator
from great_expectations.datasource.types import S3BatchKwargs
from great_expectations.exceptions import BatchKwargsError

logger = logging.getLogger(__name__)


class S3GlobReaderBatchKwargsGenerator(BatchKwargsGenerator):
    """
    S3 BatchKwargGenerator provides support for generating batches of data from an S3 bucket. For the S3 batch kwargs generator, assets must
    be individually defined using a prefix and glob, although several additional configuration parameters are available
    for assets (see below).

    Example configuration::

        datasources:
          my_datasource:
            ...
            batch_kwargs_generator:
              my_s3_generator:
                class_name: S3GlobReaderBatchKwargsGenerator
                bucket: my_bucket.my_organization.priv
                reader_method: parquet  # This will be automatically inferred from suffix where possible, but can be explicitly specified as well
                reader_options:  # Note that reader options can be specified globally or per-asset
                  sep: ","
                delimiter: "/"  # Note that this is the delimiter for the BUCKET KEYS. By default it is "/"
                boto3_options:
                  endpoint_url: $S3_ENDPOINT  # Use the S3_ENDPOINT environment variable to determine which endpoint to use
                max_keys: 100  # The maximum number of keys to fetch in a single list_objects request to s3. When accessing batch_kwargs through an iterator, the iterator will silently refetch if more keys were available
                assets:
                  my_first_asset:
                    prefix: my_first_asset/
                    regex_filter: .*  # The regex filter will filter the results returned by S3 for the key and prefix to only those matching the regex
                    dictionary_assets: True
                  access_logs:
                    prefix: access_logs
                    regex_filter: access_logs/2019.*\.csv.gz
                    sep: "~"
                    max_keys: 100
    """

    # FIXME add tests for new partitioner functionality
    def __init__(self,
                 name="default",
                 datasource=None,
                 bucket=None,
                 reader_options=None,
                 assets=None,
                 delimiter="/",
                 reader_method=None,
                 boto3_options=None,
                 max_keys=1000):
        """Initialize a new S3GlobReaderBatchKwargsGenerator

        Args:
            name: the name of the batch kwargs generator
            datasource: the datasource to which it is attached
            bucket: the name of the s3 bucket from which it generates batch_kwargs
            reader_options: options passed to the datasource reader method
            assets: asset configuration (see class docstring for more information)
            delimiter: the BUCKET KEY delimiter
            reader_method: the reader_method to include in generated batch_kwargs
            boto3_options: dictionary of key-value pairs to use when creating boto3 client or resource objects
            max_keys: the maximum number of keys to fetch in a single list_objects request to s3
        """
        super(S3GlobReaderBatchKwargsGenerator, self).__init__(name, datasource=datasource)
        if reader_options is None:
            reader_options = {}

        if assets is None:
            assets = {
                "default": {
                    "prefix": "",
                    "regex_filter": ".*"
                }
            }

        self._bucket = bucket
        self._reader_method = reader_method
        self._reader_options = reader_options
        self._assets = assets
        self._delimiter = delimiter
        if boto3_options is None:
            boto3_options = {}
        self._max_keys = max_keys
        self._iterators = {}
        try:
            self._s3 = boto3.client('s3', **boto3_options)
        except TypeError:
            raise(ImportError("Unable to load boto3, which is required for S3 batch kwargs generator"))

    @property
    def reader_options(self):
        return self._reader_options

    @property
    def assets(self):
        return self._assets

    @property
    def bucket(self):
        return self._bucket

    def get_available_data_asset_names(self):
        return {"names": [(key, "file") for key in self._assets.keys()]}


    def _get_iterator(self, generator_asset, reader_options=None, limit=None):
        logger.debug("Beginning S3GlobReaderBatchKwargsGenerator _get_iterator for generator_asset: %s" % generator_asset)

        if generator_asset not in self._assets:
            batch_kwargs = {
                "generator_asset": generator_asset,
                "reader_options": reader_options,
                "limit": limit
            }
            raise BatchKwargsError("Unknown asset_name %s" % generator_asset, batch_kwargs)

        if generator_asset not in self._iterators:
            self._iterators[generator_asset] = {}

        asset_config = self._assets[generator_asset]

        return self._build_asset_iterator(
            asset_config=asset_config,
            iterator_dict=self._iterators[generator_asset],
            reader_options=reader_options,
            limit=limit
        )

    def _build_batch_kwargs_path_iter(self, path_list, reader_options=None, limit=None):
        for path in path_list:
            yield self._build_batch_kwargs(path, reader_options=reader_options, limit=limit)

    def build_batch_kwargs_from_partition_id(self, generator_asset, partition_id=None, reader_options=None, limit=None):
        try:
            asset_config = self._assets[generator_asset]
        except KeyError:
            raise GreatExpectationsError(
                "No asset config found for asset %s" % generator_asset
            )
        if generator_asset not in self._iterators:
            self._iterators[generator_asset] = {}

        iterator_dict = self._iterators[generator_asset]
        batch_kwargs = None
        for key in self._get_asset_options(generator_asset, iterator_dict):
            if self._partitioner(key=key, asset_config=asset_config) == partition_id:
                batch_kwargs = self._build_batch_kwargs(key=key, asset_config=asset_config,
                                                        reader_options=reader_options, limit=limit)

        if batch_kwargs is None:
            raise BatchKwargsError(
                "Unable to identify partition %s for asset %s" % (partition_id, generator_asset),
               {
                    generator_asset: generator_asset,
                    partition_id: partition_id
                }
            )

        return batch_kwargs

    def _build_batch_kwargs(self, key, asset_config=None, reader_options=None, limit=None):
        batch_kwargs = {"s3": "s3a://" + self.bucket + "/" + key, 'reader_options': self.reader_options}
        if asset_config.get("reader_options"):
            batch_kwargs['reader_options'].update(asset_config.get("reader_options"))
        if reader_options is not None:
            batch_kwargs['reader_options'].update(reader_options)

        if self._reader_method is not None:
            batch_kwargs["reader_method"] = self._reader_method
        if asset_config.get("reader_method"):
            batch_kwargs['reader_method'] = asset_config.get("reader_method")

        if limit:
            batch_kwargs['limit'] = limit

        return S3BatchKwargs(batch_kwargs)

    def _get_asset_options(self, asset_config, iterator_dict):
        query_options = {
            "Bucket": self.bucket,
            "Delimiter": asset_config.get("delimiter", self._delimiter),
            "Prefix": asset_config.get("prefix", None),
            "MaxKeys": asset_config.get("max_keys", self._max_keys)
        }
        directory_assets = asset_config.get("directory_assets", False)

        if "continuation_token" in iterator_dict:
            query_options.update({
                "ContinuationToken": iterator_dict["continuation_token"]
            })

        logger.debug("Fetching objects from S3 with query options: %s" % str(query_options))
        asset_options = self._s3.list_objects_v2(**query_options)
        if directory_assets:
            if "CommonPrefixes" not in asset_options:
                raise BatchKwargsError(
                    "Unable to build batch_kwargs. The asset may not be configured correctly. If directory assets "
                    "are requested, then common prefixes must be returned.",
                    {
                        "asset_configuration": asset_config,
                        "contents": asset_options["Contents"] if "Contents" in asset_options else None
                    }
                )
            keys = [item["Prefix"] for item in asset_options["CommonPrefixes"]]
        else:
            if "Contents" not in asset_options:
                raise BatchKwargsError(
                    "Unable to build batch_kwargs. The asset may not be configured correctly. If s3 returned common "
                    "prefixes it may not have been able to identify desired keys, and they are included in the "
                    "incomplete batch_kwargs object returned with this error.",
                    {
                        "asset_configuration": asset_config,
                        "common_prefixes": asset_options["CommonPrefixes"]
                        if "CommonPrefixes" in asset_options else None
                    }
                )
            keys = [item["Key"] for item in asset_options["Contents"] if item["Size"] > 0]

        keys = [key for key in
                filter(lambda x: re.match(asset_config.get("regex_filter", ".*"), x) is not None, keys)]
        for key in keys:
            yield key

        if asset_options["IsTruncated"]:
            iterator_dict["continuation_token"] = asset_options["NextContinuationToken"]
            # Recursively fetch more
            for key in self._get_asset_options(asset_config, iterator_dict):
                yield key
        elif "continuation_token" in iterator_dict:
            # Make sure we clear the token once we've gotten fully through
            del iterator_dict["continuation_token"]

    def _build_asset_iterator(self, asset_config, iterator_dict, reader_options=None, limit=None):
        for key in self._get_asset_options(asset_config, iterator_dict):
            yield self._build_batch_kwargs(
                key,
                asset_config,
                reader_options=reader_options,
                limit=limit
            )

    def get_available_partition_ids(self, generator_asset):
        if generator_asset not in self._iterators:
            self._iterators[generator_asset] = {}
        iterator_dict = self._iterators[generator_asset]
        asset_config = self._assets[generator_asset]
        available_ids = [
            self._partitioner(key=key, asset_config=asset_config)
            for key in self._get_asset_options(generator_asset, iterator_dict)
        ]
        return available_ids

    def _partitioner(self, key, asset_config):
        if "partition_regex" in asset_config:
            match_group_id = asset_config.get("match_group_id", 1)
            matches = re.match(asset_config["partition_regex"], key)
            # In the case that there is a defined regex, the user *wanted* a partition. But it didn't match.
            # So, we'll add a *sortable* id
            if matches is None:
                logger.warning("No match found for key: %s" % key)
                return datetime.datetime.utcnow().strftime("%Y%m%dT%H%M%S.%fZ") + "__unmatched"
            else:
                try:
                    return matches.group(match_group_id)
                except IndexError:
                    logger.warning("No match group %d in key %s" % (match_group_id, key))
                    return datetime.datetime.utcnow().strftime("%Y%m%dT%H%M%S.%fZ") + "__no_match_group"

        # If there is no partitioner defined, fall back on using the path as a partition_id
        else:
            prefix = asset_config.get("prefix", "")
            if key.startswith(prefix):
                key = key[len(prefix):]
            return key




from great_expectations.core.usage_statistics.anonymizers.anonymizer import Anonymizer
from great_expectations.data_context.store import (
    Store,
    ValidationsStore,
    ExpectationsStore,
    HtmlSiteStore,
    MetricStore,
    EvaluationParameterStore,
)
from great_expectations.core.usage_statistics.anonymizers.store_backend_anonymizer import StoreBackendAnonymizer


class StoreAnonymizer(Anonymizer):
    def __init__(self, salt=None):
        super(StoreAnonymizer, self).__init__(salt=salt)
        # ordered bottom up in terms of inheritance order
        self._ge_classes = [
            ValidationsStore,
            ExpectationsStore,
            EvaluationParameterStore,
            MetricStore,
            Store,
            HtmlSiteStore
        ]
        self._store_backend_anonymizer = StoreBackendAnonymizer(salt=salt)

    def anonymize_store_info(self, store_name, store_obj):
        anonymized_info_dict = dict()
        anonymized_info_dict["anonymized_name"] = self.anonymize(store_name)
        store_backend_obj = store_obj.store_backend

        self.anonymize_object_info(
            object_=store_obj,
            anonymized_info_dict=anonymized_info_dict,
            ge_classes=self._ge_classes
        )

        anonymized_info_dict["anonymized_store_backend"] = self._store_backend_anonymizer.anonymize_store_backend_info(
            store_backend_obj=store_backend_obj
        )

        return anonymized_info_dict





import logging
import os
import random
import re
import shutil
# PYTHON 2 - py2 - update to ABC direct use rather than __metaclass__ once we drop py2 support
import logging
from abc import ABCMeta

from great_expectations.data_context.store.store_backend import StoreBackend
from great_expectations.exceptions import StoreBackendError

logger = logging.getLogger(__name__)


class TupleStoreBackend(StoreBackend, metaclass=ABCMeta):
    """
    If filepath_template is provided, the key to this StoreBackend abstract class must be a tuple with 
    fixed length equal to the number of unique components matching the regex r"{\d+}"

    For example, in the following template path: expectations/{0}/{1}/{2}/prefix-{2}.json, keys must have
    three components.
    """

    def __init__(self, filepath_template=None, filepath_prefix=None, filepath_suffix=None, forbidden_substrings=None,
                 platform_specific_separator=True, fixed_length_key=False):
        super().__init__(fixed_length_key=fixed_length_key)
        if forbidden_substrings is None:
            forbidden_substrings = ["/", "\\"]
        self.forbidden_substrings = forbidden_substrings
        self.platform_specific_separator = platform_specific_separator

        if filepath_template is not None and filepath_suffix is not None:
            raise ValueError("filepath_suffix may only be used when filepath_template is None")

        self.filepath_template = filepath_template
        if filepath_prefix and len(filepath_prefix) > 0:
            # Validate that the filepath prefix does not end with a forbidden substring
            if filepath_prefix[-1] in self.forbidden_substrings:
                raise StoreBackendError("Unable to initialize TupleStoreBackend: filepath_prefix may not end with a "
                                        "forbidden substring. Current forbidden substrings are " +
                                        str(forbidden_substrings))
        self.filepath_prefix = filepath_prefix
        self.filepath_suffix = filepath_suffix

        if filepath_template is not None:
            # key length is the number of unique values to be substituted in the filepath_template
            self.key_length = len(
                set(
                    re.findall(r"{\d+}", filepath_template)
                )
            )

            self.verify_that_key_to_filepath_operation_is_reversible()
            self._fixed_length_key = True

    def _validate_key(self, key):
        super()._validate_key(key)

        for key_element in key:
            for substring in self.forbidden_substrings:
                if substring in key_element:
                    raise ValueError("Keys in {0} must not contain substrings in {1} : {2}".format(
                        self.__class__.__name__,
                        self.forbidden_substrings,
                        key,
                    ))

    def _validate_value(self, value):
        if not isinstance(value, str) and not isinstance(value, bytes):
            raise TypeError("Values in {0} must be instances of {1} or {2}, not {3}".format(
                self.__class__.__name__,
                str,
                bytes,
                type(value),
            ))

    def _convert_key_to_filepath(self, key):
        # NOTE: This method uses a hard-coded forward slash as a separator,
        # and then replaces that with a platform-specific separator if requested (the default)
        self._validate_key(key)
        if self.filepath_template:
            converted_string = self.filepath_template.format(*list(key))
        else:
            converted_string = '/'.join(key)

        if self.filepath_prefix:
            converted_string = self.filepath_prefix + "/" + converted_string
        if self.filepath_suffix:
            converted_string += self.filepath_suffix
        if self.platform_specific_separator:
            converted_string = os.path.normpath(converted_string)

        return converted_string

    def _convert_filepath_to_key(self, filepath):
        if self.platform_specific_separator:
            filepath = os.path.normpath(filepath)

        if self.filepath_prefix:
            if not filepath.startswith(self.filepath_prefix) and len(filepath) >= len(self.filepath_prefix) + 1:
                # If filepath_prefix is set, we expect that it is the first component of a valid filepath.
                raise ValueError("filepath must start with the filepath_prefix when one is set by the store_backend")
            else:
                # Remove the prefix before processing
                # Also remove the separator that was added, which may have been platform-dependent
                filepath = filepath[len(self.filepath_prefix) + 1:]

        if self.filepath_suffix:
            if not filepath.endswith(self.filepath_suffix):
                # If filepath_suffix is set, we expect that it is the last component of a valid filepath.
                raise ValueError("filepath must end with the filepath_suffix when one is set by the store_backend")
            else:
                # Remove the suffix before processing
                filepath = filepath[:-len(self.filepath_suffix)]

        if self.filepath_template:
            # filepath_template is always specified with forward slashes, but it is then
            # used to (1) dynamically construct and evaluate a regex, and (2) split the provided (observed) filepath
            if self.platform_specific_separator:
                filepath_template = os.path.join(*self.filepath_template.split('/'))
                filepath_template = filepath_template.replace('\\', '\\\\')
            else:
                filepath_template = self.filepath_template

            # Convert the template to a regex
            indexed_string_substitutions = re.findall(r"{\d+}", filepath_template)
            tuple_index_list = ["(?P<tuple_index_{0}>.*)".format(i, ) for i in range(len(indexed_string_substitutions))]
            intermediate_filepath_regex = re.sub(
                r"{\d+}",
                lambda m, r=iter(tuple_index_list): next(r),
                filepath_template
            )
            filepath_regex = intermediate_filepath_regex.format(*tuple_index_list)

            # Apply the regex to the filepath
            matches = re.compile(filepath_regex).match(filepath)
            if matches is None:
                return None

            # Map key elements into the appropriate parts of the tuple
            new_key = [None] * self.key_length
            for i in range(len(tuple_index_list)):
                tuple_index = int(re.search(r'\d+', indexed_string_substitutions[i]).group(0))
                key_element = matches.group('tuple_index_' + str(i))
                new_key[tuple_index] = key_element

            new_key = tuple(new_key)
        else:
            filepath = os.path.normpath(filepath)
            new_key = tuple(filepath.split(os.sep))

        return new_key

    def verify_that_key_to_filepath_operation_is_reversible(self):
        def get_random_hex(size=4):
            return "".join([random.choice(list("ABCDEF0123456789")) for _ in range(size)])

        key = tuple([get_random_hex() for _ in range(self.key_length)])
        filepath = self._convert_key_to_filepath(key)
        new_key = self._convert_filepath_to_key(filepath)
        if key != new_key:
            raise ValueError(
                "filepath template {0} for class {1} is not reversible for a tuple of length {2}. "
                "Have you included all elements in the key tuple?".format(
                    self.filepath_template,
                    self.__class__.__name__,
                    self.key_length,
                ))


class TupleFilesystemStoreBackend(TupleStoreBackend):
    """Uses a local filepath as a store.

    The key to this StoreBackend must be a tuple with fixed length based on the filepath_template,
    or a variable-length tuple may be used and returned with an optional filepath_suffix (to be) added.
    The filepath_template is a string template used to convert the key to a filepath.
    """

    def __init__(self,
                 base_directory,
                 filepath_template=None,
                 filepath_prefix=None,
                 filepath_suffix=None,
                 forbidden_substrings=None,
                 platform_specific_separator=True,
                 root_directory=None,
                 fixed_length_key=False):
        super().__init__(
            filepath_template=filepath_template,
            filepath_prefix=filepath_prefix,
            filepath_suffix=filepath_suffix,
            forbidden_substrings=forbidden_substrings,
            platform_specific_separator=platform_specific_separator,
            fixed_length_key=fixed_length_key
        )
        if os.path.isabs(base_directory):
            self.full_base_directory = base_directory
        else:
            if root_directory is None:
                raise ValueError("base_directory must be an absolute path if root_directory is not provided")
            elif not os.path.isabs(root_directory):
                raise ValueError("root_directory must be an absolute path. Got {0} instead.".format(root_directory))
            else:
                self.full_base_directory = os.path.join(root_directory, base_directory)

        os.makedirs(str(os.path.dirname(self.full_base_directory)), exist_ok=True)

    def _get(self, key):
        filepath = os.path.join(
            self.full_base_directory,
            self._convert_key_to_filepath(key)
        )
        with open(filepath) as infile:
            return infile.read()

    def _set(self, key, value, **kwargs):
        if not isinstance(key, tuple):
            key = key.to_tuple()
        filepath = os.path.join(
            self.full_base_directory,
            self._convert_key_to_filepath(key)
        )
        path, filename = os.path.split(filepath)

        os.makedirs(str(path), exist_ok=True)
        with open(filepath, "wb") as outfile:
            if isinstance(value, str):
                outfile.write(value.encode("utf-8"))
            else:
                outfile.write(value)
        return filepath

    def list_keys(self, prefix=()):
        key_list = []
        for root, dirs, files in os.walk(os.path.join(self.full_base_directory, *prefix)):
            for file_ in files:
                full_path, file_name = os.path.split(os.path.join(root, file_))
                relative_path = os.path.relpath(
                    full_path,
                    self.full_base_directory,
                )
                if relative_path == ".":
                    filepath = file_name
                else:
                    filepath = os.path.join(
                        relative_path,
                        file_name
                    )

                if self.filepath_prefix and not filepath.startswith(self.filepath_prefix):
                    continue
                elif self.filepath_suffix and not filepath.endswith(self.filepath_suffix):
                    continue
                key = self._convert_filepath_to_key(filepath)
                if key and not self.is_ignored_key(key):
                    key_list.append(key)

        return key_list
    
    def remove_key(self, key):
        if not isinstance(key, tuple):
            key = key.to_tuple()

        filepath = os.path.join(
            self.full_base_directory,
            self._convert_key_to_filepath(key)
        )
        path, filename = os.path.split(filepath)
        if os.path.exists(filepath):
            if shutil.rmtree(self.full_base_directory):
                return True
        return False

    def get_url_for_key(self, key, protocol=None):
        path = self._convert_key_to_filepath(key)
        full_path = os.path.join(self.full_base_directory, path)
        if protocol is None:
            protocol = "file:"
        url = protocol + "//" + full_path

        return url

    def _has_key(self, key):
        return os.path.isfile(os.path.join(self.full_base_directory, self._convert_key_to_filepath(key)))


class TupleS3StoreBackend(TupleStoreBackend):
    """
    Uses an S3 bucket as a store.

    The key to this StoreBackend must be a tuple with fixed length based on the filepath_template,
    or a variable-length tuple may be used and returned with an optional filepath_suffix (to be) added.
    The filepath_template is a string template used to convert the key to a filepath.
    """

    def __init__(
            self,
            bucket,
            prefix="",
            filepath_template=None,
            filepath_prefix=None,
            filepath_suffix=None,
            forbidden_substrings=None,
            platform_specific_separator=False,
            fixed_length_key=False
    ):
        super().__init__(
            filepath_template=filepath_template,
            filepath_prefix=filepath_prefix,
            filepath_suffix=filepath_suffix,
            forbidden_substrings=forbidden_substrings,
            platform_specific_separator=platform_specific_separator,
            fixed_length_key=fixed_length_key
        )
        self.bucket = bucket
        self.prefix = prefix

    def _get(self, key):
        s3_object_key = os.path.join(
            self.prefix,
            self._convert_key_to_filepath(key)
        )

        import boto3
        s3 = boto3.client('s3')
        s3_response_object = s3.get_object(Bucket=self.bucket, Key=s3_object_key)
        return s3_response_object['Body'].read().decode(s3_response_object.get("ContentEncoding", 'utf-8'))

    def _set(self, key, value, content_encoding='utf-8', content_type='application/json'):
        s3_object_key = os.path.join(
            self.prefix,
            self._convert_key_to_filepath(key)
        )

        import boto3
        s3 = boto3.resource('s3')
        result_s3 = s3.Object(self.bucket, s3_object_key)
        if isinstance(value, str):
            result_s3.put(Body=value.encode(content_encoding), ContentEncoding=content_encoding,
                          ContentType=content_type)
        else:
            result_s3.put(Body=value, ContentType=content_type)
        return s3_object_key

    def list_keys(self):
        key_list = []

        import boto3
        s3 = boto3.client('s3')

        s3_objects = s3.list_objects(Bucket=self.bucket, Prefix=self.prefix)
        if "Contents" in s3_objects:
            objects = s3_objects["Contents"]
        elif "CommonPrefixes" in s3_objects:
            logger.warning("TupleS3StoreBackend returned CommonPrefixes, but delimiter should not have been set.")
            objects = []
        else:
            # No objects found in store
            objects = []

        for s3_object_info in objects:
            s3_object_key = s3_object_info['Key']
            s3_object_key = os.path.relpath(
                s3_object_key,
                self.prefix,
            )
            if self.filepath_prefix and not s3_object_key.startswith(self.filepath_prefix):
                continue
            elif self.filepath_suffix and not s3_object_key.endswith(self.filepath_suffix):
                continue
            key = self._convert_filepath_to_key(s3_object_key)
            if key:
                key_list.append(key)

        return key_list

    def get_url_for_key(self, key, protocol=None):
        import boto3

        location = boto3.client('s3').get_bucket_location(Bucket=self.bucket)['LocationConstraint']
        if location is None:
            location = "s3"
        else:
            location = "s3-" + location
        s3_key = self._convert_key_to_filepath(key)
        if not self.prefix:
            return f"https://{location}.amazonaws.com/{self.bucket}/{s3_key}"
        return f"https://{location}.amazonaws.com/{self.bucket}/{self.prefix}/{s3_key}"

    def remove_key(self, key):
        import boto3
        from botocore.exceptions import ClientError
        s3 = boto3.resource('s3')
        s3_key = self._convert_key_to_filepath(key)
        if s3_key:
            try:
                #s3.Object(boto3.client('s3').get_bucket_location(Bucket=self.bucket), s3_key).delete()
                objects_to_delete = s3.meta.client.list_objects(Bucket=self.bucket, Prefix=self.prefix)

                delete_keys = {'Objects' : []}
                delete_keys['Objects'] = [{'Key' : k} for k in [obj['Key'] for obj in objects_to_delete.get('Contents', [])]]
                s3.meta.client.delete_objects(Bucket=self.bucket, Delete=delete_keys)
                return True
            except ClientError as e:
                return False
        else:
            return False

    def _has_key(self, key):
        all_keys = self.list_keys()
        return key in all_keys


class TupleGCSStoreBackend(TupleStoreBackend):
    """
    Uses a GCS bucket as a store.

    The key to this StoreBackend must be a tuple with fixed length based on the filepath_template,
    or a variable-length tuple may be used and returned with an optional filepath_suffix (to be) added.

    The filepath_template is a string template used to convert the key to a filepath.
    """

    def __init__(
            self,
            bucket,
            prefix,
            project,
            filepath_template=None,
            filepath_prefix=None,
            filepath_suffix=None,
            forbidden_substrings=None,
            platform_specific_separator=False,
            fixed_length_key=False
    ):
        super().__init__(
            filepath_template=filepath_template,
            filepath_prefix=filepath_prefix,
            filepath_suffix=filepath_suffix,
            forbidden_substrings=forbidden_substrings,
            platform_specific_separator=platform_specific_separator,
            fixed_length_key=fixed_length_key
        )
        self.bucket = bucket
        self.prefix = prefix
        self.project = project

    def _get(self, key):
        gcs_object_key = os.path.join(
            self.prefix,
            self._convert_key_to_filepath(key)
        )

        from google.cloud import storage
        gcs = storage.Client(project=self.project)
        bucket = gcs.get_bucket(self.bucket)
        gcs_response_object = bucket.get_blob(gcs_object_key)
        return gcs_response_object.download_as_string().decode("utf-8")

    def _set(self, key, value, content_encoding='utf-8', content_type='application/json'):
        gcs_object_key = os.path.join(
            self.prefix,
            self._convert_key_to_filepath(key)
        )

        from google.cloud import storage
        gcs = storage.Client(project=self.project)
        bucket = gcs.get_bucket(self.bucket)
        blob = bucket.blob(gcs_object_key)
        if isinstance(value, str):
            blob.upload_from_string(value.encode(content_encoding), content_encoding=content_encoding,
                                    content_type=content_type)
        else:
            blob.upload_from_string(value, content_type=content_type)
        return gcs_object_key

    def list_keys(self):
        key_list = []

        from google.cloud import storage
        gcs = storage.Client(self.project)

        for blob in gcs.list_blobs(self.bucket, prefix=self.prefix):
            gcs_object_name = blob.name
            gcs_object_key = os.path.relpath(
                gcs_object_name,
                self.prefix,
            )
            if self.filepath_prefix and not gcs_object_key.startswith(self.filepath_prefix):
                continue
            elif self.filepath_suffix and not gcs_object_key.endswith(self.filepath_suffix):
                continue
            key = self._convert_filepath_to_key(gcs_object_key)
            if key:
                key_list.append(key)

        return key_list

    def get_url_for_key(self, key, protocol=None):
        path = self._convert_key_to_filepath(key)
        return "https://storage.googleapis.com/" + self.bucket + "/" + path

    def remove_key(self, key):
        from google.cloud import storage
        from gcloud.exceptions import NotFound
        gcs = storage.Client(project=self.project)
        bucket = gcs.get_bucket(self.bucket)
        try:
            bucket.delete_blobs(blobs=bucket.list_blobs(prefix=self.prefix))
        except NotFound:
            return False
        return True

    def _has_key(self, key):
        all_keys = self.list_keys()
        return key in all_keys



from great_expectations.core import ExpectationSuiteValidationResult

try:
    from unittest import mock
except ImportError:
    import mock

from great_expectations.validation_operators import (
    SlackNotificationAction,
    StoreValidationResultAction
)
from great_expectations.data_context.store import (
    ValidationsStore,
)
from great_expectations.data_context.types.resource_identifiers import (
    ValidationResultIdentifier,
    ExpectationSuiteIdentifier
)


def test_StoreAction():
    fake_in_memory_store = ValidationsStore(
        store_backend={
            "class_name": "InMemoryStoreBackend",
        }
    )
    stores = {
        "fake_in_memory_store": fake_in_memory_store
    }

    class Object(object):
        pass

    data_context = Object()
    data_context.stores = stores

    action = StoreValidationResultAction(
        data_context=data_context,
        target_store_name="fake_in_memory_store",
    )
    assert fake_in_memory_store.list_keys() == []

    action.run(
        validation_result_suite_identifier=ValidationResultIdentifier(
            expectation_suite_identifier=ExpectationSuiteIdentifier(
                expectation_suite_name="default_expectations"
            ),
            run_id="prod_20190801",
            batch_identifier="1234"
        ),
        validation_result_suite=ExpectationSuiteValidationResult(
            success=False,
            results=[]
        ),
        data_asset=None
    )

    assert len(fake_in_memory_store.list_keys()) == 1
    stored_identifier = fake_in_memory_store.list_keys()[0]
    assert stored_identifier.batch_identifier == "1234"
    assert stored_identifier.expectation_suite_identifier.expectation_suite_name == "default_expectations"
    assert stored_identifier.run_id == "prod_20190801"

    assert fake_in_memory_store.get(ValidationResultIdentifier(
        expectation_suite_identifier=ExpectationSuiteIdentifier(
            expectation_suite_name="default_expectations"
        ),
        run_id="prod_20190801",
        batch_identifier="1234"
    )) == ExpectationSuiteValidationResult(
        success=False,
        results=[]
    )


def test_SlackNotificationAction(data_context):
    renderer = {
        "module_name": "great_expectations.render.renderer.slack_renderer",
        "class_name": "SlackRenderer",
    }
    slack_webhook = "https://hooks.slack.com/services/test/slack/webhook"
    notify_on = "all"

    slack_action = SlackNotificationAction(
        data_context=data_context,
        renderer=renderer,
        slack_webhook=slack_webhook,
        notify_on=notify_on
    )

    validation_result_suite = ExpectationSuiteValidationResult(results=[], success=True,
                                                               statistics={'evaluated_expectations': 0,
                                                                           'successful_expectations': 0,
                                                                           'unsuccessful_expectations': 0,
                                                                           'success_percent': None},
                                                               meta={
                                                                   'great_expectations.__version__': 'v0.8.0__develop',
                                                                   'expectation_suite_name': 'asset.default',
                                                                   'run_id': 'test_100'})

    validation_result_suite_id = ValidationResultIdentifier(expectation_suite_identifier=ExpectationSuiteIdentifier(
        "asset.default"), run_id="test_100", batch_identifier="1234")

    # TODO: improve this test - currently it is verifying a failed call to Slack
    assert slack_action.run(
        validation_result_suite_identifier=validation_result_suite_id,
        validation_result_suite=validation_result_suite,
        data_asset=None
    ) == None

# def test_ExtractAndStoreEvaluationParamsAction():
#     fake_in_memory_store = ValidationsStore(
#         root_directory = None,
#         store_backend = {
#             "class_name": "InMemoryStoreBackend",
#         }
#     )
#     stores = {
#         "fake_in_memory_store" : fake_in_memory_store
#     }
#
#     # NOTE: This is a hack meant to last until we implement runtime_environments
#     class Object(object):
#         pass
#
#     data_context = Object()
#     data_context.stores = stores
#
#     action = StoreValidationResultAction(
#         data_context = data_context,
#         target_store_name = "fake_in_memory_store",
#     )
#     assert fake_in_memory_store.list_keys() == []
#
#     vr_id = "ValidationResultIdentifier.my_db.default_generator.my_table.default_expectations.prod_20190801"
#     action.run(
#         validation_result_suite_id=ValidationResultIdentifier(from_string=vr_id),
#         validation_result_suite={},
#         data_asset=None
#     )
#
#     assert len(fake_in_memory_store.list_keys()) == 1
#     assert fake_in_memory_store.list_keys()[0].to_string() == "ValidationResultIdentifier.my_db.default_generator.my_table.default_expectations.prod_20190801"
#     assert fake_in_memory_store.get(ValidationResultIdentifier(
#         from_string="ValidationResultIdentifier.my_db.default_generator.my_table.default_expectations.prod_20190801"
#     )) == {}
#



from great_expectations.util import verify_dynamic_loading_support
from great_expectations.core import ExpectationSuiteValidationResultSchema
from great_expectations.data_context.store.database_store_backend import DatabaseStoreBackend
from great_expectations.data_context.store.store import Store
from great_expectations.data_context.store.tuple_store_backend import TupleStoreBackend
from great_expectations.data_context.types.resource_identifiers import ValidationResultIdentifier
from great_expectations.data_context.util import load_class


class ValidationsStore(Store):
    _key_class = ValidationResultIdentifier

    def __init__(self, store_backend=None, runtime_environment=None):
        self._expectationSuiteValidationResultSchema = ExpectationSuiteValidationResultSchema()

        if store_backend is not None:
            store_backend_module_name = store_backend.get("module_name", "great_expectations.data_context.store")
            store_backend_class_name = store_backend.get("class_name", "InMemoryStoreBackend")
            verify_dynamic_loading_support(module_name=store_backend_module_name)
            store_backend_class = load_class(store_backend_class_name, store_backend_module_name)

            # Store Backend Class was loaded successfully; verify that it is of a correct subclass.
            if issubclass(store_backend_class, TupleStoreBackend):
                # Provide defaults for this common case
                store_backend["filepath_suffix"] = store_backend.get("filepath_suffix", ".json")
            elif issubclass(store_backend_class, DatabaseStoreBackend):
                # Provide defaults for this common case
                store_backend["table_name"] = store_backend.get("table_name", "ge_validations_store")
                store_backend["key_columns"] = store_backend.get(
                    "key_columns", [
                        "expectation_suite_name",
                        "run_id",
                        "batch_identifier"
                    ]
                )
        super().__init__(store_backend=store_backend, runtime_environment=runtime_environment)

    def serialize(self, key, value):
        return self._expectationSuiteValidationResultSchema.dumps(value)

    def deserialize(self, key, value):
        return self._expectationSuiteValidationResultSchema.loads(value)
